#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from nfs3.x on Sun Oct 04 08:09:27 2015
"""
NFSv3 decoding module
"""
import nfstest_config as c
import nfs3_const as const
from packet.utils import *
from baseobj import BaseObj
from packet.unpack import Unpack

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "3.0"

#
# Constants
class nfs_bool(Enum):
    """enum nfs_bool"""
    _enumdict = const.nfs_bool

# Basic data types
uint64      = Unpack.unpack_uint64
int64       = Unpack.unpack_int64
uint32      = Unpack.unpack_uint
int32       = Unpack.unpack_int
filename3   = Unpack.unpack_opaque
nfspath3    = Unpack.unpack_opaque
fileid3     = uint64
cookie3     = uint64
cookieverf3 = lambda unpack: StrHex(unpack.unpack_fopaque(const.NFS3_COOKIEVERFSIZE))
createverf3 = lambda unpack: StrHex(unpack.unpack_fopaque(const.NFS3_CREATEVERFSIZE))
writeverf3  = lambda unpack: StrHex(unpack.unpack_fopaque(const.NFS3_WRITEVERFSIZE))
uid3        = uint32
gid3        = uint32
size3       = uint64
offset3     = uint64
mode3       = uint32
count3      = uint32
nfs_fh3     = lambda unpack: StrHex(unpack.unpack_opaque(const.NFS3_FHSIZE))
access3     = uint32

# Error status
class nfsstat3(Enum):
    """enum nfsstat3"""
    _enumdict = const.nfsstat3

class ftype3(Enum):
    """enum ftype3"""
    _enumdict = const.ftype3

class specdata3(BaseObj):
    """
       struct specdata3 {
           uint32 specdata1;
           uint32 specdata2;
       };
    """
    # Class attributes
    _strfmt1  = "major:{0} minor:{1}"
    _attrlist = ("specdata1", "specdata2")

    def __init__(self, unpack):
        self.specdata1 = uint32(unpack)
        self.specdata2 = uint32(unpack)

class nfstime3(BaseObj):
    """
       struct nfstime3 {
           uint32 seconds;
           uint32 nseconds;
       };
    """
    # Class attributes
    _strfmt1  = "{0}.{1:09}"
    _attrlist = ("seconds", "nseconds")

    def __init__(self, unpack):
        self.seconds  = uint32(unpack)
        self.nseconds = uint32(unpack)

class fattr3(BaseObj):
    """
       struct fattr3 {
           ftype3    type;
           mode3     mode;
           uint32    nlink;
           uid3      uid;
           gid3      gid;
           size3     size;
           size3     used;
           specdata3 rdev;
           uint64    fsid;
           fileid3   fileid;
           nfstime3  atime;
           nfstime3  mtime;
           nfstime3  ctime;
       };
    """
    # Class attributes
    _strfmt1  = "{0} mode:{1:04o} nlink:{2} uid:{3} gid:{4} size:{5} fileid:{9}"
    _attrlist = ("type", "mode", "nlink", "uid", "gid", "size", "used", "rdev",
                 "fsid", "fileid", "atime", "mtime", "ctime")

    def __init__(self, unpack):
        self.type   = ftype3(unpack)
        self.mode   = mode3(unpack)
        self.nlink  = uint32(unpack)
        self.uid    = uid3(unpack)
        self.gid    = gid3(unpack)
        self.size   = size3(unpack)
        self.used   = size3(unpack)
        self.rdev   = specdata3(unpack)
        self.fsid   = uint64(unpack)
        self.fileid = fileid3(unpack)
        self.atime  = nfstime3(unpack)
        self.mtime  = nfstime3(unpack)
        self.ctime  = nfstime3(unpack)

class post_op_attr(BaseObj):
    """
       union switch post_op_attr (bool attributes_follow) {
           case const.TRUE:
               fattr3 attributes;
           case const.FALSE:
               void;
       };
    """
    def __init__(self, unpack):
        self.set_attr("attributes_follow", nfs_bool(unpack))
        if self.attributes_follow == const.TRUE:
            self.set_attr("attributes", fattr3(unpack), switch=True)

class wcc_attr(BaseObj):
    """
       struct wcc_attr {
           size3    size;
           nfstime3 mtime;
           nfstime3 ctime;
       };
    """
    # Class attributes
    _attrlist = ("size", "mtime", "ctime")

    def __init__(self, unpack):
        self.size  = size3(unpack)
        self.mtime = nfstime3(unpack)
        self.ctime = nfstime3(unpack)

class pre_op_attr(BaseObj):
    """
       union switch pre_op_attr (bool attributes_follow) {
           case const.TRUE:
               wcc_attr attributes;
           case const.FALSE:
               void;
       };
    """
    def __init__(self, unpack):
        self.set_attr("attributes_follow", nfs_bool(unpack))
        if self.attributes_follow == const.TRUE:
            self.set_attr("attributes", wcc_attr(unpack), switch=True)

class wcc_data(BaseObj):
    """
       struct wcc_data {
           pre_op_attr  before;
           post_op_attr after;
       };
    """
    # Class attributes
    _attrlist = ("before", "after")

    def __init__(self, unpack):
        self.before = pre_op_attr(unpack)
        self.after  = post_op_attr(unpack)

class post_op_fh3(BaseObj):
    """
       union switch post_op_fh3 (bool handle_follows) {
           case const.TRUE:
               nfs_fh3 fh;
           case const.FALSE:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "FH:{1:crc32}"

    def __init__(self, unpack):
        self.set_attr("handle_follows", nfs_bool(unpack))
        if self.handle_follows == const.TRUE:
            self.set_attr("fh", nfs_fh3(unpack), switch=True)
        elif self.handle_follows == const.FALSE:
            self.set_strfmt(1, "")

class time_how(Enum):
    """enum time_how"""
    _enumdict = const.time_how

class set_mode3(BaseObj):
    """
       union switch set_mode3 (bool set_it) {
           case const.TRUE:
               mode3 mode;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "mode:{1:04o}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", nfs_bool(unpack))
        if self.set_it == const.TRUE:
            self.set_attr("mode", mode3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class set_uid3(BaseObj):
    """
       union switch set_uid3 (bool set_it) {
           case const.TRUE:
               uid3 uid;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "uid:{1}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", nfs_bool(unpack))
        if self.set_it == const.TRUE:
            self.set_attr("uid", uid3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class set_gid3(BaseObj):
    """
       union switch set_gid3 (bool set_it) {
           case const.TRUE:
               gid3 gid;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "gid:{1}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", nfs_bool(unpack))
        if self.set_it == const.TRUE:
            self.set_attr("gid", gid3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class set_size3(BaseObj):
    """
       union switch set_size3 (bool set_it) {
           case const.TRUE:
               size3 size;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "size:{1}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", nfs_bool(unpack))
        if self.set_it == const.TRUE:
            self.set_attr("size", size3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class set_atime(BaseObj):
    """
       union switch set_atime (time_how set_it) {
           case const.SET_TO_CLIENT_TIME:
               nfstime3 atime;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "atime:{1}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", time_how(unpack))
        if self.set_it == const.SET_TO_CLIENT_TIME:
            self.set_attr("atime", nfstime3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class set_mtime(BaseObj):
    """
       union switch set_mtime (time_how set_it) {
           case const.SET_TO_CLIENT_TIME:
               nfstime3 mtime;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "mtime:{1}\x20"

    def __init__(self, unpack):
        self.set_attr("set_it", time_how(unpack))
        if self.set_it == const.SET_TO_CLIENT_TIME:
            self.set_attr("mtime", nfstime3(unpack), switch=True)
        else:
            self.set_strfmt(1, "")

class sattr3(BaseObj):
    """
       struct sattr3 {
           set_mode3 mode;
           set_uid3  uid;
           set_gid3  gid;
           set_size3 size;
           set_atime atime;
           set_mtime mtime;
       };
    """
    # Class attributes
    _strfmt1  = "{0}{1}{2}{3}"
    _attrlist = ("mode", "uid", "gid", "size", "atime", "mtime")

    def __init__(self, unpack):
        self.mode  = set_mode3(unpack)
        self.uid   = set_uid3(unpack)
        self.gid   = set_gid3(unpack)
        self.size  = set_size3(unpack)
        self.atime = set_atime(unpack)
        self.mtime = set_mtime(unpack)

class diropargs3(BaseObj):
    """
       struct diropargs3 {
           nfs_fh3   fh;
           filename3 name;
       };
    """
    # Class attributes
    _strfmt1  = "DH:{0:crc32}/{1}"
    _attrlist = ("fh", "name")

    def __init__(self, unpack):
        self.fh   = nfs_fh3(unpack)
        self.name = filename3(unpack)

# GETATTR3res NFSPROC3_GETATTR(GETATTR3args) = 1;
class GETATTR3args(BaseObj):
    """
       struct GETATTR3args {
           nfs_fh3 fh;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh",)

    def __init__(self, unpack):
        self.fh = nfs_fh3(unpack)

class GETATTR3resok(BaseObj):
    """
       struct GETATTR3resok {
           fattr3 attributes;
       };
    """
    # Class attributes
    _strfmt1  = "{0}"
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = fattr3(unpack)

class GETATTR3res(BaseObj):
    """
       union switch GETATTR3res (nfsstat3 status) {
           case const.NFS3_OK:
               GETATTR3resok resok;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", GETATTR3resok(unpack), switch=True)

# SETATTR3res NFSPROC3_SETATTR(SETATTR3args) = 2;
class sattrguard3(BaseObj):
    """
       union switch sattrguard3 (bool check) {
           case const.TRUE:
               nfstime3 ctime;
           case const.FALSE:
               void;
       };
    """
    def __init__(self, unpack):
        self.set_attr("check", nfs_bool(unpack))
        if self.check == const.TRUE:
            self.set_attr("ctime", nfstime3(unpack), switch=True)

class SETATTR3args(BaseObj):
    """
       struct SETATTR3args {
           nfs_fh3     fh;
           sattr3      attributes;
           sattrguard3 guard;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} {1}"
    _attrlist = ("fh", "attributes", "guard")

    def __init__(self, unpack):
        self.fh         = nfs_fh3(unpack)
        self.attributes = sattr3(unpack)
        self.guard      = sattrguard3(unpack)

class SETATTR3resok(BaseObj):
    """
       struct SETATTR3resok {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class SETATTR3resfail(BaseObj):
    """
       struct SETATTR3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class SETATTR3res(BaseObj):
    """
       union switch SETATTR3res (nfsstat3 status) {
           case const.NFS3_OK:
               SETATTR3resok resok;
           default:
               SETATTR3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", SETATTR3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", SETATTR3resfail(unpack), switch=True)

# LOOKUP3res NFSPROC3_LOOKUP(LOOKUP3args) = 3;
class LOOKUP3args(BaseObj):
    """
       struct LOOKUP3args {
           diropargs3 what;
       };
    """
    # Class attributes
    _fattrs   = ("what",)
    _strfmt1  = "{0}"
    _attrlist = ("what",)

    def __init__(self, unpack):
        self.what = diropargs3(unpack)

class LOOKUP3resok(BaseObj):
    """
       struct LOOKUP3resok {
           nfs_fh3      fh;
           post_op_attr attributes;
           post_op_attr dir_attributes;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh", "attributes", "dir_attributes")

    def __init__(self, unpack):
        self.fh             = nfs_fh3(unpack)
        self.attributes     = post_op_attr(unpack)
        self.dir_attributes = post_op_attr(unpack)

class LOOKUP3resfail(BaseObj):
    """
       struct LOOKUP3resfail {
           post_op_attr dir_attributes;
       };
    """
    # Class attributes
    _attrlist = ("dir_attributes",)

    def __init__(self, unpack):
        self.dir_attributes = post_op_attr(unpack)

class LOOKUP3res(BaseObj):
    """
       union switch LOOKUP3res (nfsstat3 status) {
           case const.NFS3_OK:
               LOOKUP3resok resok;
           default:
               LOOKUP3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", LOOKUP3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", LOOKUP3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

class ACCESS3args(BaseObj):
    """
       struct ACCESS3args {
           nfs_fh3 fh;
           access3 access;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} acc:{1:#04x}"
    _attrlist = ("fh", "access")

    def __init__(self, unpack):
        self.fh     = nfs_fh3(unpack)
        self.access = access3(unpack)

class ACCESS3resok(BaseObj):
    """
       struct ACCESS3resok {
           post_op_attr attributes;
           access3      access;
       };
    """
    # Class attributes
    _strfmt1  = "acc:{1:#04x}"
    _attrlist = ("attributes", "access")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.access     = access3(unpack)

class ACCESS3resfail(BaseObj):
    """
       struct ACCESS3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class ACCESS3res(BaseObj):
    """
       union switch ACCESS3res (nfsstat3 status) {
           case const.NFS3_OK:
               ACCESS3resok resok;
           default:
               ACCESS3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", ACCESS3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", ACCESS3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# READLINK3res NFSPROC3_READLINK(READLINK3args) = 5;
class READLINK3args(BaseObj):
    """
       struct READLINK3args {
           nfs_fh3 fh;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh",)

    def __init__(self, unpack):
        self.fh = nfs_fh3(unpack)

class READLINK3resok(BaseObj):
    """
       struct READLINK3resok {
           post_op_attr attributes;
           nfspath3     link;
       };
    """
    # Class attributes
    _strfmt1  = "{1}"
    _attrlist = ("attributes", "link")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.link       = nfspath3(unpack)

class READLINK3resfail(BaseObj):
    """
       struct READLINK3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class READLINK3res(BaseObj):
    """
       union switch READLINK3res (nfsstat3 status) {
           case const.NFS3_OK:
               READLINK3resok resok;
           default:
               READLINK3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", READLINK3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", READLINK3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# READ3res NFSPROC3_READ(READ3args) = 6;
class READ3args(BaseObj):
    """
       struct READ3args {
           nfs_fh3 fh;
           offset3 offset;
           count3  count;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} off:{1:umax64} len:{2:umax32}"
    _attrlist = ("fh", "offset", "count")

    def __init__(self, unpack):
        self.fh     = nfs_fh3(unpack)
        self.offset = offset3(unpack)
        self.count  = count3(unpack)

class READ3resok(BaseObj):
    """
       struct READ3resok {
           post_op_attr attributes;
           count3       count;
           bool         eof;
           opaque       data<>;
       };
    """
    # Class attributes
    _strfmt1  = "eof:{2} count:{1:umax32}"
    _attrlist = ("attributes", "count", "eof", "data")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.count      = count3(unpack)
        self.eof        = nfs_bool(unpack)
        self.data       = unpack.unpack_opaque()

class READ3resfail(BaseObj):
    """
       struct READ3resfail {
           post_op_attr file_attributes;
       };
    """
    # Class attributes
    _attrlist = ("file_attributes",)

    def __init__(self, unpack):
        self.file_attributes = post_op_attr(unpack)

class READ3res(BaseObj):
    """
       union switch READ3res (nfsstat3 status) {
           case const.NFS3_OK:
               READ3resok resok;
           default:
               READ3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", READ3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", READ3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# WRITE3res NFSPROC3_WRITE(WRITE3args) = 7;
class stable_how(Enum):
    """enum stable_how"""
    _enumdict = const.stable_how

class WRITE3args(BaseObj):
    """
       struct WRITE3args {
           nfs_fh3    fh;
           offset3    offset;
           count3     count;
           stable_how stable;
           opaque     data<>;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} off:{1:umax64} len:{2:umax32} {3}"
    _attrlist = ("fh", "offset", "count", "stable", "data")

    def __init__(self, unpack):
        self.fh     = nfs_fh3(unpack)
        self.offset = offset3(unpack)
        self.count  = count3(unpack)
        self.stable = stable_how(unpack)
        self.data   = unpack.unpack_opaque()

class WRITE3resok(BaseObj):
    """
       struct WRITE3resok {
           wcc_data   wcc;
           count3     count;
           stable_how committed;
           writeverf3 verifier;
       };
    """
    # Class attributes
    _strfmt1  = "count:{1:umax32} verf:{3} {2}"
    _attrlist = ("wcc", "count", "committed", "verifier")

    def __init__(self, unpack):
        self.wcc       = wcc_data(unpack)
        self.count     = count3(unpack)
        self.committed = stable_how(unpack)
        self.verifier  = writeverf3(unpack)

class WRITE3resfail(BaseObj):
    """
       struct WRITE3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class WRITE3res(BaseObj):
    """
       union switch WRITE3res (nfsstat3 status) {
           case const.NFS3_OK:
               WRITE3resok resok;
           default:
               WRITE3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", WRITE3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", WRITE3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# CREATE3res NFSPROC3_CREATE(CREATE3args) = 8;
class createmode3(Enum):
    """enum createmode3"""
    _enumdict = const.createmode3

class createhow3(BaseObj):
    """
       union switch createhow3 (createmode3 mode) {
           case const.UNCHECKED:
           case const.GUARDED:
               sattr3 attributes;
           case const.EXCLUSIVE:
               createverf3 verifier;
       };
    """
    # Class attributes
    _strfmt1 = "{0}"

    def __init__(self, unpack):
        self.set_attr("mode", createmode3(unpack))
        if self.mode in [const.UNCHECKED, const.GUARDED]:
            self.set_attr("attributes", sattr3(unpack), switch=True)
        elif self.mode == const.EXCLUSIVE:
            self.set_attr("verifier", createverf3(unpack), switch=True)

class CREATE3args(BaseObj):
    """
       struct CREATE3args {
           diropargs3 where;
           createhow3 how;
       };
    """
    # Class attributes
    _fattrs   = ("where",)
    _strfmt1  = "{0} {1}"
    _attrlist = ("where", "how")

    def __init__(self, unpack):
        self.where = diropargs3(unpack)
        self.how   = createhow3(unpack)

class CREATE3resok(BaseObj):
    """
       struct CREATE3resok {
           post_op_fh3  obj;
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _fattrs   = ("obj",)
    _strfmt1  = "{0}"
    _attrlist = ("obj", "attributes", "wcc")

    def __init__(self, unpack):
        self.obj        = post_op_fh3(unpack)
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class CREATE3resfail(BaseObj):
    """
       struct CREATE3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class CREATE3res(BaseObj):
    """
       union switch CREATE3res (nfsstat3 status) {
           case const.NFS3_OK:
               CREATE3resok resok;
           default:
               CREATE3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", CREATE3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", CREATE3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# MKDIR3res NFSPROC3_MKDIR(MKDIR3args) = 9;
class MKDIR3args(BaseObj):
    """
       struct MKDIR3args {
           diropargs3 where;
           sattr3     attributes;
       };
    """
    # Class attributes
    _fattrs   = ("where",)
    _strfmt1  = "{0} {1}"
    _attrlist = ("where", "attributes")

    def __init__(self, unpack):
        self.where      = diropargs3(unpack)
        self.attributes = sattr3(unpack)

class MKDIR3resok(BaseObj):
    """
       struct MKDIR3resok {
           post_op_fh3  obj;
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _fattrs   = ("obj",)
    _strfmt1  = "{0}"
    _attrlist = ("obj", "attributes", "wcc")

    def __init__(self, unpack):
        self.obj        = post_op_fh3(unpack)
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class MKDIR3resfail(BaseObj):
    """
       struct MKDIR3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class MKDIR3res(BaseObj):
    """
       union switch MKDIR3res (nfsstat3 status) {
           case const.NFS3_OK:
               MKDIR3resok resok;
           default:
               MKDIR3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", MKDIR3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", MKDIR3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# SYMLINK3res NFSPROC3_SYMLINK(SYMLINK3args) = 10;
class symlinkdata3(BaseObj):
    """
       struct symlinkdata3 {
           sattr3   attributes;
           nfspath3 linkdata;
       };
    """
    # Class attributes
    _strfmt1  = "{1} {0}"
    _attrlist = ("attributes", "linkdata")

    def __init__(self, unpack):
        self.attributes = sattr3(unpack)
        self.linkdata   = nfspath3(unpack)

class SYMLINK3args(BaseObj):
    """
       struct SYMLINK3args {
           diropargs3   where;
           symlinkdata3 symlink;
       };
    """
    # Class attributes
    _fattrs   = ("where",)
    _strfmt1  = "{0} -> {1}"
    _attrlist = ("where", "symlink")

    def __init__(self, unpack):
        self.where   = diropargs3(unpack)
        self.symlink = symlinkdata3(unpack)

class SYMLINK3resok(BaseObj):
    """
       struct SYMLINK3resok {
           post_op_fh3  obj;
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _fattrs   = ("obj",)
    _strfmt1  = "{0}"
    _attrlist = ("obj", "attributes", "wcc")

    def __init__(self, unpack):
        self.obj        = post_op_fh3(unpack)
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class SYMLINK3resfail(BaseObj):
    """
       struct SYMLINK3resfail {
           wcc_data dir_wcc;
       };
    """
    # Class attributes
    _attrlist = ("dir_wcc",)

    def __init__(self, unpack):
        self.dir_wcc = wcc_data(unpack)

class SYMLINK3res(BaseObj):
    """
       union switch SYMLINK3res (nfsstat3 status) {
           case const.NFS3_OK:
               SYMLINK3resok resok;
           default:
               SYMLINK3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", SYMLINK3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", SYMLINK3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# MKNOD3res NFSPROC3_MKNOD(MKNOD3args) = 11;
class devicedata3(BaseObj):
    """
       struct devicedata3 {
           sattr3    attributes;
           specdata3 spec;
       };
    """
    # Class attributes
    _strfmt1  = "{0} {1}"
    _attrlist = ("attributes", "spec")

    def __init__(self, unpack):
        self.attributes = sattr3(unpack)
        self.spec       = specdata3(unpack)

class mknoddata3(BaseObj):
    """
       union switch mknoddata3 (ftype3 type) {
           case const.NF3CHR:
           case const.NF3BLK:
               devicedata3 device;
           case const.NF3SOCK:
           case const.NF3FIFO:
               sattr3 attributes;
           default:
               void;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("type", ftype3(unpack))
        if self.type in [const.NF3CHR, const.NF3BLK]:
            self.set_attr("device", devicedata3(unpack), switch=True)
            self.set_strfmt(1, "{1}")
        elif self.type in [const.NF3SOCK, const.NF3FIFO]:
            self.set_attr("attributes", sattr3(unpack), switch=True)

class MKNOD3args(BaseObj):
    """
       struct MKNOD3args {
           diropargs3 where;
           mknoddata3 what;
       };
    """
    # Class attributes
    _fattrs   = ("where",)
    _strfmt1  = "{1.type} {0} {1}"
    _attrlist = ("where", "what")

    def __init__(self, unpack):
        self.where = diropargs3(unpack)
        self.what  = mknoddata3(unpack)

class MKNOD3resok(BaseObj):
    """
       struct MKNOD3resok {
           post_op_fh3  obj;
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _fattrs   = ("obj",)
    _strfmt1  = "{0}"
    _attrlist = ("obj", "attributes", "wcc")

    def __init__(self, unpack):
        self.obj        = post_op_fh3(unpack)
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class MKNOD3resfail(BaseObj):
    """
       struct MKNOD3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class MKNOD3res(BaseObj):
    """
       union switch MKNOD3res (nfsstat3 status) {
           case const.NFS3_OK:
               MKNOD3resok resok;
           default:
               MKNOD3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", MKNOD3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", MKNOD3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# REMOVE3res NFSPROC3_REMOVE(REMOVE3args) = 12;
class REMOVE3args(BaseObj):
    """
       struct REMOVE3args {
           diropargs3 object;
       };
    """
    # Class attributes
    _fattrs   = ("object",)
    _strfmt1  = "{0}"
    _attrlist = ("object",)

    def __init__(self, unpack):
        self.object = diropargs3(unpack)

class REMOVE3resok(BaseObj):
    """
       struct REMOVE3resok {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class REMOVE3resfail(BaseObj):
    """
       struct REMOVE3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class REMOVE3res(BaseObj):
    """
       union switch REMOVE3res (nfsstat3 status) {
           case const.NFS3_OK:
               REMOVE3resok resok;
           default:
               REMOVE3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", REMOVE3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", REMOVE3resfail(unpack), switch=True)

# RMDIR3res NFSPROC3_RMDIR(RMDIR3args) = 13;
class RMDIR3args(BaseObj):
    """
       struct RMDIR3args {
           diropargs3 object;
       };
    """
    # Class attributes
    _fattrs   = ("object",)
    _strfmt1  = "{0}"
    _attrlist = ("object",)

    def __init__(self, unpack):
        self.object = diropargs3(unpack)

class RMDIR3resok(BaseObj):
    """
       struct RMDIR3resok {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class RMDIR3resfail(BaseObj):
    """
       struct RMDIR3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class RMDIR3res(BaseObj):
    """
       union switch RMDIR3res (nfsstat3 status) {
           case const.NFS3_OK:
               RMDIR3resok resok;
           default:
               RMDIR3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", RMDIR3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", RMDIR3resfail(unpack), switch=True)

# RENAME3res NFSPROC3_RENAME(RENAME3args) = 14;
class RENAME3args(BaseObj):
    """
       struct RENAME3args {
           diropargs3 nfrom;
           diropargs3 nto;
       };
    """
    # Class attributes
    _fattrs   = ("nfrom",)
    _strfmt1  = "{0} -> {1}"
    _attrlist = ("nfrom", "nto")

    def __init__(self, unpack):
        self.nfrom   = diropargs3(unpack)
        self.nto     = diropargs3(unpack)
        self.newname = self.nto.name

class RENAME3resok(BaseObj):
    """
       struct RENAME3resok {
           wcc_data fromdir_wcc;
           wcc_data todir_wcc;
       };
    """
    # Class attributes
    _attrlist = ("fromdir_wcc", "todir_wcc")

    def __init__(self, unpack):
        self.fromdir_wcc = wcc_data(unpack)
        self.todir_wcc   = wcc_data(unpack)

class RENAME3resfail(BaseObj):
    """
       struct RENAME3resfail {
           wcc_data fromdir_wcc;
           wcc_data todir_wcc;
       };
    """
    # Class attributes
    _attrlist = ("fromdir_wcc", "todir_wcc")

    def __init__(self, unpack):
        self.fromdir_wcc = wcc_data(unpack)
        self.todir_wcc   = wcc_data(unpack)

class RENAME3res(BaseObj):
    """
       union switch RENAME3res (nfsstat3 status) {
           case const.NFS3_OK:
               RENAME3resok resok;
           default:
               RENAME3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", RENAME3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", RENAME3resfail(unpack), switch=True)

# LINK3res NFSPROC3_LINK(LINK3args) = 15;
class LINK3args(BaseObj):
    """
       struct LINK3args {
           nfs_fh3    fh;
           diropargs3 link;
       };
    """
    # Class attributes
    _fattrs   = ("link",)
    _strfmt1  = "{1} -> FH:{0:crc32}"
    _attrlist = ("fh", "link")

    def __init__(self, unpack):
        self.fh   = nfs_fh3(unpack)
        self.link = diropargs3(unpack)

class LINK3resok(BaseObj):
    """
       struct LINK3resok {
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _attrlist = ("attributes", "wcc")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class LINK3resfail(BaseObj):
    """
       struct LINK3resfail {
           post_op_attr attributes;
           wcc_data     wcc;
       };
    """
    # Class attributes
    _attrlist = ("attributes", "wcc")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.wcc        = wcc_data(unpack)

class LINK3res(BaseObj):
    """
       union switch LINK3res (nfsstat3 status) {
           case const.NFS3_OK:
               LINK3resok resok;
           default:
               LINK3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", LINK3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", LINK3resfail(unpack), switch=True)

# READDIR3res NFSPROC3_READDIR(READDIR3args) = 16;
class READDIR3args(BaseObj):
    """
       struct READDIR3args {
           nfs_fh3     fh;
           cookie3     cookie;
           cookieverf3 verifier;
           count3      count;
       };
    """
    # Class attributes
    _strfmt1  = "DH:{0:crc32} cookie:{1} verf:{2} count:{3:umax32}"
    _attrlist = ("fh", "cookie", "verifier", "count")

    def __init__(self, unpack):
        self.fh       = nfs_fh3(unpack)
        self.cookie   = cookie3(unpack)
        self.verifier = cookieverf3(unpack)
        self.count    = count3(unpack)

class entry3(BaseObj):
    """
       struct entry3 {
           fileid3   fileid;
           filename3 name;
           cookie3   cookie;
           entry3    *nextentry;
       };
    """
    # Class attributes
    _attrlist = ("fileid", "name", "cookie")

    def __init__(self, unpack):
        self.fileid = fileid3(unpack)
        self.name   = filename3(unpack)
        self.cookie = cookie3(unpack)

class dirlist3(BaseObj):
    """
       struct dirlist3 {
           entry3 *entries;
           bool   eof;
       };
    """
    # Class attributes
    _strfmt1  = "eof:{1}"
    _attrlist = ("entries", "eof")

    def __init__(self, unpack):
        try:
            self.entries = unpack.unpack_list(entry3)
            self.eof     = nfs_bool(unpack)
        except:
            pass

class READDIR3resok(BaseObj):
    """
       struct READDIR3resok {
           post_op_attr attributes;
           cookieverf3  verifier;
           dirlist3     reply;
       };
    """
    # Class attributes
    _fattrs   = ("reply",)
    _strfmt1  = "verf:{1} {2}"
    _attrlist = ("attributes", "verifier", "reply")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.verifier   = cookieverf3(unpack)
        self.reply      = dirlist3(unpack)

class READDIR3resfail(BaseObj):
    """
       struct READDIR3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class READDIR3res(BaseObj):
    """
       union switch READDIR3res (nfsstat3 status) {
           case const.NFS3_OK:
               READDIR3resok resok;
           default:
               READDIR3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", READDIR3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", READDIR3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# READDIRPLUS3res NFSPROC3_READDIRPLUS(READDIRPLUS3args) = 17;
class READDIRPLUS3args(BaseObj):
    """
       struct READDIRPLUS3args {
           nfs_fh3     fh;
           cookie3     cookie;
           cookieverf3 verifier;
           count3      dircount;
           count3      maxcount;
       };
    """
    # Class attributes
    _strfmt1  = "DH:{0:crc32} cookie:{1} verf:{2} count:{3:umax32}"
    _attrlist = ("fh", "cookie", "verifier", "dircount", "maxcount")

    def __init__(self, unpack):
        self.fh       = nfs_fh3(unpack)
        self.cookie   = cookie3(unpack)
        self.verifier = cookieverf3(unpack)
        self.dircount = count3(unpack)
        self.maxcount = count3(unpack)

class entryplus3(BaseObj):
    """
       struct entryplus3 {
           fileid3      fileid;
           filename3    name;
           cookie3      cookie;
           post_op_attr attributes;
           post_op_fh3  obj;
           entryplus3   *nextentry;
       };
    """
    # Class attributes
    _fattrs   = ("obj",)
    _attrlist = ("fileid", "name", "cookie", "attributes", "obj")

    def __init__(self, unpack):
        self.fileid     = fileid3(unpack)
        self.name       = filename3(unpack)
        self.cookie     = cookie3(unpack)
        self.attributes = post_op_attr(unpack)
        self.obj        = post_op_fh3(unpack)

class dirlistplus3(BaseObj):
    """
       struct dirlistplus3 {
           entryplus3 *entries;
           bool       eof;
       };
    """
    # Class attributes
    _strfmt1  = "eof:{1}"
    _attrlist = ("entries", "eof")

    def __init__(self, unpack):
        try:
            self.entries = unpack.unpack_list(entryplus3)
            self.eof     = nfs_bool(unpack)
        except:
            pass

class READDIRPLUS3resok(BaseObj):
    """
       struct READDIRPLUS3resok {
           post_op_attr attributes;
           cookieverf3  verifier;
           dirlistplus3 reply;
       };
    """
    # Class attributes
    _fattrs   = ("reply",)
    _strfmt1  = "verf:{1} {2}"
    _attrlist = ("attributes", "verifier", "reply")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.verifier   = cookieverf3(unpack)
        self.reply      = dirlistplus3(unpack)

class READDIRPLUS3resfail(BaseObj):
    """
       struct READDIRPLUS3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class READDIRPLUS3res(BaseObj):
    """
       union switch READDIRPLUS3res (nfsstat3 status) {
           case const.NFS3_OK:
               READDIRPLUS3resok resok;
           default:
               READDIRPLUS3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", READDIRPLUS3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", READDIRPLUS3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# FSSTAT3res NFSPROC3_FSSTAT(FSSTAT3args) = 18;
class FSSTAT3args(BaseObj):
    """
       struct FSSTAT3args {
           nfs_fh3 fh;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh",)

    def __init__(self, unpack):
        self.fh = nfs_fh3(unpack)

class FSSTAT3resok(BaseObj):
    """
       struct FSSTAT3resok {
           post_op_attr attributes;
           size3        tbytes;
           size3        fbytes;
           size3        abytes;
           size3        tfiles;
           size3        ffiles;
           size3        afiles;
           uint32       invarsec;
       };
    """
    # Class attributes
    _attrlist = ("attributes", "tbytes", "fbytes", "abytes", "tfiles", "ffiles",
                 "afiles", "invarsec")

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)
        self.tbytes     = size3(unpack)
        self.fbytes     = size3(unpack)
        self.abytes     = size3(unpack)
        self.tfiles     = size3(unpack)
        self.ffiles     = size3(unpack)
        self.afiles     = size3(unpack)
        self.invarsec   = uint32(unpack)

class FSSTAT3resfail(BaseObj):
    """
       struct FSSTAT3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class FSSTAT3res(BaseObj):
    """
       union switch FSSTAT3res (nfsstat3 status) {
           case const.NFS3_OK:
               FSSTAT3resok resok;
           default:
               FSSTAT3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", FSSTAT3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", FSSTAT3resfail(unpack), switch=True)

class FSINFO3args(BaseObj):
    """
       struct FSINFO3args {
           nfs_fh3 fh;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh",)

    def __init__(self, unpack):
        self.fh = nfs_fh3(unpack)

class FSINFO3resok(BaseObj):
    """
       struct FSINFO3resok {
           post_op_attr attributes;
           uint32       rtmax;
           uint32       rtpref;
           uint32       rtmult;
           uint32       wtmax;
           uint32       wtpref;
           uint32       wtmult;
           uint32       dtpref;
           size3        maxfilesize;
           nfstime3     time_delta;
           uint32       properties;
       };
    """
    # Class attributes
    _attrlist = ("attributes", "rtmax", "rtpref", "rtmult", "wtmax", "wtpref",
                 "wtmult", "dtpref", "maxfilesize", "time_delta", "properties")

    def __init__(self, unpack):
        self.attributes  = post_op_attr(unpack)
        self.rtmax       = uint32(unpack)
        self.rtpref      = uint32(unpack)
        self.rtmult      = uint32(unpack)
        self.wtmax       = uint32(unpack)
        self.wtpref      = uint32(unpack)
        self.wtmult      = uint32(unpack)
        self.dtpref      = uint32(unpack)
        self.maxfilesize = size3(unpack)
        self.time_delta  = nfstime3(unpack)
        self.properties  = uint32(unpack)

class FSINFO3resfail(BaseObj):
    """
       struct FSINFO3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class FSINFO3res(BaseObj):
    """
       union switch FSINFO3res (nfsstat3 status) {
           case const.NFS3_OK:
               FSINFO3resok resok;
           default:
               FSINFO3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", FSINFO3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", FSINFO3resfail(unpack), switch=True)

# PATHCONF3res NFSPROC3_PATHCONF(PATHCONF3args) = 20;
class PATHCONF3args(BaseObj):
    """
       struct PATHCONF3args {
           nfs_fh3 fh;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32}"
    _attrlist = ("fh",)

    def __init__(self, unpack):
        self.fh = nfs_fh3(unpack)

class PATHCONF3resok(BaseObj):
    """
       struct PATHCONF3resok {
           post_op_attr attributes;
           uint32       linkmax;
           uint32       name_max;
           bool         no_trunc;
           bool         chown_restricted;
           bool         case_insensitive;
           bool         case_preserving;
       };
    """
    # Class attributes
    _attrlist = ("attributes", "linkmax", "name_max", "no_trunc",
                 "chown_restricted", "case_insensitive", "case_preserving")

    def __init__(self, unpack):
        self.attributes       = post_op_attr(unpack)
        self.linkmax          = uint32(unpack)
        self.name_max         = uint32(unpack)
        self.no_trunc         = nfs_bool(unpack)
        self.chown_restricted = nfs_bool(unpack)
        self.case_insensitive = nfs_bool(unpack)
        self.case_preserving  = nfs_bool(unpack)

class PATHCONF3resfail(BaseObj):
    """
       struct PATHCONF3resfail {
           post_op_attr attributes;
       };
    """
    # Class attributes
    _attrlist = ("attributes",)

    def __init__(self, unpack):
        self.attributes = post_op_attr(unpack)

class PATHCONF3res(BaseObj):
    """
       union switch PATHCONF3res (nfsstat3 status) {
           case const.NFS3_OK:
               PATHCONF3resok resok;
           default:
               PATHCONF3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = ""

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", PATHCONF3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", PATHCONF3resfail(unpack), switch=True)

# COMMIT3res NFSPROC3_COMMIT(COMMIT3args) = 21;
class COMMIT3args(BaseObj):
    """
       struct COMMIT3args {
           nfs_fh3 fh;
           offset3 offset;
           count3  count;
       };
    """
    # Class attributes
    _strfmt1  = "FH:{0:crc32} off:{1:umax64} len:{2:umax32}"
    _attrlist = ("fh", "offset", "count")

    def __init__(self, unpack):
        self.fh     = nfs_fh3(unpack)
        self.offset = offset3(unpack)
        self.count  = count3(unpack)

class COMMIT3resok(BaseObj):
    """
       struct COMMIT3resok {
           wcc_data   wcc;
           writeverf3 verifier;
       };
    """
    # Class attributes
    _strfmt1  = "verf:{1}"
    _attrlist = ("wcc", "verifier")

    def __init__(self, unpack):
        self.wcc      = wcc_data(unpack)
        self.verifier = writeverf3(unpack)

class COMMIT3resfail(BaseObj):
    """
       struct COMMIT3resfail {
           wcc_data wcc;
       };
    """
    # Class attributes
    _attrlist = ("wcc",)

    def __init__(self, unpack):
        self.wcc = wcc_data(unpack)

class COMMIT3res(BaseObj):
    """
       union switch COMMIT3res (nfsstat3 status) {
           case const.NFS3_OK:
               COMMIT3resok resok;
           default:
               COMMIT3resfail resfail;
       };
    """
    # Class attributes
    _strfmt1 = "{1}"

    def __init__(self, unpack):
        self.set_attr("status", nfsstat3(unpack))
        if self.status == const.NFS3_OK:
            self.set_attr("resok", COMMIT3resok(unpack), switch=True)
        else:
            self.set_attr("resfail", COMMIT3resfail(unpack), switch=True)
            self.set_strfmt(1, "")

# Procedures
class nfs_proc3(Enum):
    """enum nfs_proc3"""
    _enumdict = const.nfs_proc3

class NFS3args(RPCload):
    """
       union switch NFS3args (nfs_proc3 procedure) {
           case const.NFSPROC3_NULL:
               void;
           case const.NFSPROC3_GETATTR:
               GETATTR3args opgetattr;
           case const.NFSPROC3_SETATTR:
               SETATTR3args opsetattr;
           case const.NFSPROC3_LOOKUP:
               LOOKUP3args oplookup;
           case const.NFSPROC3_ACCESS:
               ACCESS3args opaccess;
           case const.NFSPROC3_READLINK:
               READLINK3args opreadlink;
           case const.NFSPROC3_READ:
               READ3args opread;
           case const.NFSPROC3_WRITE:
               WRITE3args opwrite;
           case const.NFSPROC3_CREATE:
               CREATE3args opcreate;
           case const.NFSPROC3_MKDIR:
               MKDIR3args opmkdir;
           case const.NFSPROC3_SYMLINK:
               SYMLINK3args opsymlink;
           case const.NFSPROC3_MKNOD:
               MKNOD3args opmknod;
           case const.NFSPROC3_REMOVE:
               REMOVE3args opremove;
           case const.NFSPROC3_RMDIR:
               RMDIR3args oprmdir;
           case const.NFSPROC3_RENAME:
               RENAME3args oprename;
           case const.NFSPROC3_LINK:
               LINK3args oplink;
           case const.NFSPROC3_READDIR:
               READDIR3args opreaddir;
           case const.NFSPROC3_READDIRPLUS:
               READDIRPLUS3args opreaddirplus;
           case const.NFSPROC3_FSSTAT:
               FSSTAT3args opfsstat;
           case const.NFSPROC3_FSINFO:
               FSINFO3args opfsinfo;
           case const.NFSPROC3_PATHCONF:
               PATHCONF3args oppathconf;
           case const.NFSPROC3_COMMIT:
               COMMIT3args opcommit;
       };
    """
    # Class attributes
    _pindex  = 9
    _strname = "NFS"

    def __init__(self, unpack, procedure):
        self.set_attr("procedure", nfs_proc3(procedure))
        if self.procedure == const.NFSPROC3_GETATTR:
            self.set_attr("opgetattr", GETATTR3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_SETATTR:
            self.set_attr("opsetattr", SETATTR3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_LOOKUP:
            self.set_attr("oplookup", LOOKUP3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_ACCESS:
            self.set_attr("opaccess", ACCESS3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READLINK:
            self.set_attr("opreadlink", READLINK3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READ:
            self.set_attr("opread", READ3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_WRITE:
            self.set_attr("opwrite", WRITE3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_CREATE:
            self.set_attr("opcreate", CREATE3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_MKDIR:
            self.set_attr("opmkdir", MKDIR3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_SYMLINK:
            self.set_attr("opsymlink", SYMLINK3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_MKNOD:
            self.set_attr("opmknod", MKNOD3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_REMOVE:
            self.set_attr("opremove", REMOVE3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_RMDIR:
            self.set_attr("oprmdir", RMDIR3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_RENAME:
            self.set_attr("oprename", RENAME3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_LINK:
            self.set_attr("oplink", LINK3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READDIR:
            self.set_attr("opreaddir", READDIR3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READDIRPLUS:
            self.set_attr("opreaddirplus", READDIRPLUS3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_FSSTAT:
            self.set_attr("opfsstat", FSSTAT3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_FSINFO:
            self.set_attr("opfsinfo", FSINFO3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_PATHCONF:
            self.set_attr("oppathconf", PATHCONF3args(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_COMMIT:
            self.set_attr("opcommit", COMMIT3args(unpack), switch=True)
        self.argop = self.procedure
        self.op    = self.procedure

class NFS3res(RPCload):
    """
       union switch NFS3res (nfs_proc3 procedure) {
           case const.NFSPROC3_NULL:
               void;
           case const.NFSPROC3_GETATTR:
               GETATTR3res opgetattr;
           case const.NFSPROC3_SETATTR:
               SETATTR3res opsetattr;
           case const.NFSPROC3_LOOKUP:
               LOOKUP3res oplookup;
           case const.NFSPROC3_ACCESS:
               ACCESS3res opaccess;
           case const.NFSPROC3_READLINK:
               READLINK3res opreadlink;
           case const.NFSPROC3_READ:
               READ3res opread;
           case const.NFSPROC3_WRITE:
               WRITE3res opwrite;
           case const.NFSPROC3_CREATE:
               CREATE3res opcreate;
           case const.NFSPROC3_MKDIR:
               MKDIR3res opmkdir;
           case const.NFSPROC3_SYMLINK:
               SYMLINK3res opsymlink;
           case const.NFSPROC3_MKNOD:
               MKNOD3res opmknod;
           case const.NFSPROC3_REMOVE:
               REMOVE3res opremove;
           case const.NFSPROC3_RMDIR:
               RMDIR3res oprmdir;
           case const.NFSPROC3_RENAME:
               RENAME3res oprename;
           case const.NFSPROC3_LINK:
               LINK3res oplink;
           case const.NFSPROC3_READDIR:
               READDIR3res opreaddir;
           case const.NFSPROC3_READDIRPLUS:
               READDIRPLUS3res opreaddirplus;
           case const.NFSPROC3_FSSTAT:
               FSSTAT3res opfsstat;
           case const.NFSPROC3_FSINFO:
               FSINFO3res opfsinfo;
           case const.NFSPROC3_PATHCONF:
               PATHCONF3res oppathconf;
           case const.NFSPROC3_COMMIT:
               COMMIT3res opcommit;
       };
    """
    # Class attributes
    _pindex  = 9
    _strname = "NFS"

    def __init__(self, unpack, procedure):
        self.set_attr("procedure", nfs_proc3(procedure))
        if self.procedure == const.NFSPROC3_GETATTR:
            self.set_attr("opgetattr", GETATTR3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_SETATTR:
            self.set_attr("opsetattr", SETATTR3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_LOOKUP:
            self.set_attr("oplookup", LOOKUP3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_ACCESS:
            self.set_attr("opaccess", ACCESS3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READLINK:
            self.set_attr("opreadlink", READLINK3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READ:
            self.set_attr("opread", READ3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_WRITE:
            self.set_attr("opwrite", WRITE3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_CREATE:
            self.set_attr("opcreate", CREATE3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_MKDIR:
            self.set_attr("opmkdir", MKDIR3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_SYMLINK:
            self.set_attr("opsymlink", SYMLINK3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_MKNOD:
            self.set_attr("opmknod", MKNOD3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_REMOVE:
            self.set_attr("opremove", REMOVE3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_RMDIR:
            self.set_attr("oprmdir", RMDIR3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_RENAME:
            self.set_attr("oprename", RENAME3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_LINK:
            self.set_attr("oplink", LINK3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READDIR:
            self.set_attr("opreaddir", READDIR3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_READDIRPLUS:
            self.set_attr("opreaddirplus", READDIRPLUS3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_FSSTAT:
            self.set_attr("opfsstat", FSSTAT3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_FSINFO:
            self.set_attr("opfsinfo", FSINFO3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_PATHCONF:
            self.set_attr("oppathconf", PATHCONF3res(unpack), switch=True)
        elif self.procedure == const.NFSPROC3_COMMIT:
            self.set_attr("opcommit", COMMIT3res(unpack), switch=True)
        self.resop = self.procedure
        self.op    = self.procedure



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
String Formatter object

Object used to format base objects into strings. It extends the functionality
of the string Formatter object to include new modifiers for different objects.
Some of these new modifiers include conversion of strings into a sequence
of hex characters, conversion of strings to their corresponding CRC32 or
CRC16 representation.
"""
import re
import time
import binascii
import nfstest_config as c
from string import Formatter

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.5"

# Display variables
CRC16 = True
CRC32 = True

# Maximum integer map
_max_map = {
    "max32":{
                0x7fffffff:  "max32",
               -0x80000000: "-max32",
    },
    "umax32":{
                0xffffffff: "umax32",
    },
    "max64":{
        0x7fffffffffffffff:  "max64",
       -0x8000000000000000: "-max64",
    },
    "umax64":{
        0xffffffffffffffff: "umax64",
    },
}

# Unit modifiers
UNIT_NAME = 0
UNIT_BYTE = "B"
UNIT_SEP  = ""

# Short name unit suffixes
UNIT_SUFFIXES = ["","K","M","G","T","P","E","Z"]
# Long name unit suffixes
UNIT_SUFFIX_NAME = ["", "Kilo", "Mega", "Giga", "Tera", "Peta", "Exa", "Zetta"]

def str_units(value, precision=2):
    """Convert number to a string value with units

       value:
           Number to convert
       precision:
           Return string value with the following floating point
           precision. By default no trailing zeros are returned
           but if the precision is given as a negative number
           the precision is enforced [default: 2]
    """
    # Get index to unit name
    idx = 0
    while value >= 1024:
        idx += 1
        value = value/1024.0

    if precision > 0 and round(value,precision) == int(value):
        # Remove trailing zeros when value is exact or within precision limits
        precision = 0
    if UNIT_NAME:
        suffix = UNIT_SUFFIX_NAME[idx]
    else:
        suffix = UNIT_SUFFIXES[idx]
    if len(suffix):
        suffix += UNIT_BYTE
    return "%.*f%s%s" % (abs(precision), value, UNIT_SEP, suffix)

def int_units(value):
    """Convert string value with units to an integer

       value:
           String to convert

       Examples:
           out = num_units("1MB") # out = 1048576
    """
    if type(value) == str:
        v, m = re.search(r"([-\+\.\d]+)\s*(\w?)", value).groups()
        value = int(float(v) * (1<<(10*UNIT_SUFFIXES.index(m.upper()))))
    return value

def str_time(value):
    """Convert the number of seconds to a string with a format of "[h:]mm:ss"

       value:
           Time value to convert (in seconds)

       Examples:
           out = str_time(123.0) # out = "02:03"
           out = str_time(12345) # out = "3:25:45"
    """
    ret = ""
    value = int(value)
    hh = value/3600
    mm = (value-3600*hh)/60
    ss = value%60
    if hh > 0:
        ret += "%d:" % hh
    return ret + "%02d:%02d" % (mm, ss)

def crc32(value):
    """Convert string to its crc32 representation"""
    return binascii.crc32(value) & 0xffffffff

def crc16(value):
    """Convert string to its crc16 representation"""
    return binascii.crc_hqx(value, 0xa5a5) & 0xffff

def hexstr(value):
    """Convert string to its hex representation"""
    return "0x" + value.encode("hex")

class FormatStr(Formatter):
    """String Formatter object

       FormatStr() -> New string formatter object

       Usage:
           from formatstr import FormatStr

           x = FormatStr()

           out = x.format(fmt_spec, *args, **kwargs)
           out = x.vformat(fmt_spec, args, kwargs)

           Arguments should be surrounded by curly braces {}, anything that is
           not contained in curly braces is considered literal text which is
           copied unchanged to the output.
           Positional arguments to be used in the format spec are specified
           by their index: {0}, {1}, etc.
           Named arguments to be used in the format spec are specified by
           their name: {name1}, {name2}, etc.

           Modifiers are specified after the positional index or name preceded
           by a ":", "{0:#x}" -- display first positional argument in hex

       Examples:
           # Format string using positional arguments
           out = x.format("{0} -> {1}", a, b)

           # Format string using named arguments
           out = x.format("{key}: {value}", key="id", value=32)

           # Format string using both positional and named arguments
           out = x.format("{key}: {value}, {0}, {1}", a, b, key="id", value=32)

           # Use vformat() method instead when positional arguments are given
           # as a list and named arguments are given as a dictionary
           # The following examples show the same as above
           pos_args = [a, b]
           named_args = {"key":"id", "value":32}
           out = x.vformat("{0} -> {1}", pos_args)
           out = x.vformat("{key}: {value}", named_args)
           out = x.vformat("{key}: {value}, {0}, {1}", pos_args, named_args)

           # Display string in hex
           out = x.format("{0:x}", "hello")  # out = "68656c6c6f"

           # Display string in hex with leading 0x
           out = x.format("{0:#x}", "hello") # out = "0x68656c6c6f"

           # Display string in crc32
           out = x.format("{0:crc32}", "hello") # out = "0x3610a686"

           # Display string in crc16
           out = x.format("{0:crc16}", "hello") # out = "0x9c62"

           # Substring using "@" format modifier
           # Format {0:@sindex[,eindex]} is like value[sindex:eindex]
           #   {0:@3} is like value[3:]
           #   {0:@3,5} is like value[3:5]
           #   {0:.5} is like value[:5]
           out = x.format("{0:@3}", "hello") # out = "lo"
           out = x.format("{0:.2}", "hello") # out = "he"

           # Conditionally display the first format if argument is not None,
           # else the second format is displayed
           # Format: {0:?format1:format2}
           out = x.format("{0:?tuple({0}, {1})}", 1, 2)    # out = "tuple(1, 2)"
           out = x.format("{0:?tuple({0}, {1})}", None, 2) # out = ""
           # Using 'else' format (including the escaping of else character):
           out = x.format("{0:?sid\:{0}:NONE}", 5)    # out = "sid:5"
           out = x.format("{0:?sid\:{0}:NONE}", None) # out = "NONE"

           # Nested formatting for strings, where processing is done in
           # reversed order -- process the last format first
           # Format: {0:fmtN:...:fmt2:fmt1}
           #   Display substring of 4 bytes as hex (substring then hex)
           out = x.format("{0:#x:.4}", "hello") # out = "0x68656c6c"
           #   Display first 4 bytes of string in hex (hex then substring)
           out = x.format("{0:.4:#x}", "hello") # out = "0x68"

           # Integer extension to display umax name instead of the value
           # Format: {0:max32|umax32|max64|umax64}
           # Output: if value matches the largest number in format given,
           #         the max name is displayed, else the value is displayed
           out = x.format("{0:max32}", 0x7fffffff) # out = "max32"
           out = x.format("{0:max32}", 35)         # out = "35"

           # Number extension to display the value with units
           # Format: {0:units[.precision]}
           # Output: display value as a string with units, by default
           #         precision=2 and all trailing zeros are removed.
           #         To force the precision use a negative number.
           out = x.format("{0:units}", 1024)    # out = "1KB"
           out = x.format("{0:units.4}", 2000)  # out = "1.9531KB"
           out = x.format("{0:units.-2}", 1024) # out = "1.00KB"

           # Date extension for int, long or float
           # Format: {0:date[:datefmt]}
           #         The spec given by datefmt is converted using strftime()
           #         The conversion spec "%q" is used to display microseconds
           # Output: display value as a date
           stime = 1416846041.521868
           out = x.format("{0:date}", stime) # out = "Mon Nov 24 09:20:41 2014"
           out = x.format("{0:date:%Y-%m-%d}", stime) # out = "2014-11-24"

           # List format specification
           # Format: {0[[:listfmt]:itemfmt]}
           #   If one format spec, it is applied to each item in the list
           #   If two format specs, the first is the item separator and
           #   the second is the spec applied to each item in the list
           alist = [1, 2, 3, 0xffffffff]
           out = x.format("{0:umax32}", alist)    # out = "[1, 2, 3, umax32]"
           out = x.format("{0:--:umax32}", alist) # out = "1--2--3--umax32"
    """
    def format_field(self, value, format_spec):
        """Override original method to include modifier extensions"""
        if len(format_spec) > 1 and format_spec[0] == "?":
            # Conditional directive
            # Format {0:?format1:format2}
            data = re.split(r"(?<!\\):", format_spec)
            if value is not None:
                return data[0][1:].replace("\\:", ":")
            elif len(data) > 1:
                return data[1].replace("\\:", ":")
        if value is None:
            # No value is given
            return ""
        # Process format spec
        match = re.search(r"([#@]?)(\d*)(.*)", format_spec)
        xmod, num, fmt = match.groups()
        if isinstance(value, int) and type(value) != int:
            # This is an object derived from int, convert it to string
            value = str(value)
        if isinstance(value, str):
            fmtlist = (xmod+fmt).split(":")
            if len(fmtlist) > 1:
                # Nested format, process in reversed order
                for sfmt in reversed(fmtlist):
                    value = self.format_field(value, sfmt)
                return value
            if fmt == "x":
                # Display string in hex
                xprefix = ""
                if xmod == "#":
                    xprefix = "0x"
                return xprefix + value.encode("hex")
            elif fmt == "crc32":
                if CRC32:
                    return "{0:#010x}".format(crc32(value))
                else:
                    return str(value)
            elif fmt == "crc16":
                if CRC16:
                    return "{0:#06x}".format(crc16(value))
                else:
                    return str(value)
            elif xmod == "@":
                # Format {0:@starindex[,endindex]} is like value[starindex:endindex]
                #   {0:@3} is like value[3:]
                #   {0:@3,5} is like value[3:5]
                #   {0:.5} is like value[:5]
                end = 0
                if len(fmt) > 2 and fmt[0] == ",":
                    end = int(fmt[1:])
                    return value[int(num):end]
                else:
                    return value[int(num):]
        elif isinstance(value, list):
            # Format: {0[[:listfmt]:itemfmt]}
            fmts = format_spec.split(":", 1)
            ifmt = "{0:" + fmts[-1] + "}"
            vlist = [self.format(ifmt, x) for x in value]
            if len(fmts) == 2:
                # Two format specs, use the first one for the list itself
                # and the second spec is for each item in the list
                return fmts[0].join(vlist)

            # Only one format spec is given, display list with format spec
            # applied to each item in the list
            return "[" + ", ".join(vlist) + "]"
        elif isinstance(value, int) or isinstance(value, long) or isinstance(value, float):
            if _max_map.get(fmt):
                # Format: {0:max32|umax32|max64|umax64}
                # Output: if value matches the largest number in format given,
                #         the max name is displayed, else the value is displayed
                #         {0:max32}: value:0x7fffffff then "max32" is displayed
                #         {0:max32}: value:35 then 35 is displayed
                return _max_map[fmt].get(value, str(value))
            elif fmt[:5] == "units":
                # Format: {0:units[.precision]}
                # Output: convert value to a string with units
                #         (default precision is 2)
                #         {0:units}: value:1024 then "1KB" is displayed
                #         {0:units}: value:2000 then "1.95KB is displayed
                fmts = fmt.split(".", 1)
                uargs = {}
                if len(fmts) == 2:
                    uargs["precision"] = int(fmts[1])
                return str_units(value, **uargs)
            elif fmt[:4] == "date":
                # Format: {0:date[:datefmt]}
                # Output: display value as a date
                #         value: 1416846041.521868
                #         display: 'Mon Nov 24 09:20:41 2014'
                dfmt = "%c" # Default date spec when datefmt is not given
                fmts = fmt.split(":", 1)
                if len(fmts) == 2:
                    dfmt = fmts[1]
                    if dfmt.find("%q"):
                        # Replace all instances of %q with the microseconds
                        usec = "%06d" % (1000000 * (value - int(value)))
                        dfmt = dfmt.replace("%q", usec)
                return time.strftime(dfmt, time.localtime(value))
        return format(value, format_spec)

    def get_value(self, key, args, kwargs):
        """Override original method to return "" when the positional argument
           or named argument does not exist:
             x.format("0:{0}, 1:{1}, arg1:{arg1}, arg2:{arg2}", a, arg1=11)
             the {1} will return "" since there is only one positional argument
             the {arg2} will return "" since arg2 is not a named argument
        """
        try:
            return super(FormatStr, self).get_value(key, args, kwargs)
        except (IndexError, KeyError):
            return ""



#===============================================================================
# Copyright 2016 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
NTP module

Decode NTP layer.

RFC 1059 Network Time Protocol (Version 1)
RFC 1119 Network Time Protocol (Version 2)
RFC 1305 Network Time Protocol (Version 3)
RFC 5905 Network Time Protocol (Version 4)
"""
import nfstest_config as c
from packet.utils import *
from baseobj import BaseObj

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2016 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

UINT16     = 0xffff
UINT32     = 0xffffffff
UNIX_EPOCH = 2208988800

class ntp4_mode(Enum):
    """enum ntp4_mode"""
    _enumdict = {1 : "sym_active", 2 : "sym_passive", 3 : "client",
                 4 : "server",     5 : "broadcast",   6 : "NTP_cntl"}

class NTPExtField(BaseObj):
    """NTP extension field"""
    # Class attributes
    _attrlist = ("ftype", "length", "value")

    def __init__(self, unpack):
        """Constructor which takes the Unpack object as input"""
        ulist = unpack.unpack(4, "!HH")
        self.ftype  = ulist[0]
        self.length = ulist[1]
        self.value  = unpack.read(self.length-4)

class NTP_TimeStamp(DateStr):
    """NTP timestamp"""
    _strfmt = "{0:date:%Y-%m-%d %H:%M:%S.%q}"

def ntp_timestamp(unpack):
    """Get NTP timestamp"""
    secs, fraction = unpack.unpack(8, "!II")
    if secs > 0:
        secs = secs - UNIX_EPOCH
    secs += float(fraction)/UINT32
    return NTP_TimeStamp(secs)

class NTP4(BaseObj):
    """NTP4 object

       Usage:
           from packet.application.ntp4 import NTP4

           # Decode NTP4 layer
           x = NTP4(pktt)

       Object definition:

       NTP4(
           leap       = int,    # Leap Indicator
           version    = int,    # NTP version
           mode       = int,    # Leap Indicator
           stratum    = int,    # Packet Stratum
           poll       = int,    # Maximum interval between successive messages
           precision  = float,  # Precision of system clock
           delay      = float,  # Root delay
           dispersion = float,  # Root dispersion
           refid      = string, # Reference ID
           tstamp     = float,  # Reference timestamp
           org_tstamp = float,  # Origin timestamp
           rec_tstamp = float,  # Receive timestamp
           xmt_tstamp = float,  # Transit timestamp
           fields     = list,   # Extension fields
           keyid      = int,    # Key identifier
           digest     = string, # Message digest
       )
    """
    # Class attributes
    _strfmt1  = "NTP{1} {2} {12}"
    _attrlist = ("leap", "version", "mode", "stratum", "poll", "precision",
                 "delay", "dispersion", "refid", "tstamp", "org_tstamp",
                 "rec_tstamp", "xmt_tstamp", "fields", "keyid", "digest")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        unpack = pktt.unpack
        ulist = unpack.unpack(16, "!BBbbHHHH4s")
        self.leap       = (ulist[0]>>6)&0x3
        self.version    = (ulist[0]>>3)&0x7
        self.mode       = ntp4_mode(ulist[0]&0x7)
        self.stratum    = ulist[1]
        self.poll       = 2**ulist[2]
        self.precision  = 2**ulist[3]
        self.delay      = ulist[4] + float(ulist[5])/UINT16
        self.dispersion = ulist[6] + float(ulist[7])/UINT16
        self.refid      = ulist[8]
        self.tstamp     = ntp_timestamp(unpack)
        self.org_tstamp = ntp_timestamp(unpack)
        self.rec_tstamp = ntp_timestamp(unpack)
        self.xmt_tstamp = ntp_timestamp(unpack)
        self.fields     = []
        self.keyid      = 0
        self.digest     = ""

        if self.version == 4:
            # Only NTP version 4 has extension fields
            while len(unpack) > 24:
                self.fields.append(NTPExtField(unpack))

        if self.version == 4 and len(unpack) == 20:
            # Digest is 16 bytes for NTP version 4
            self.keyid, self.digest = unpack.unpack(20, "!I16s")
        elif self.version in [2,3] and len(unpack) == 12:
            # Digest is 8 bytes for NTP version 2 and 3
            self.keyid, self.digest = unpack.unpack(12, "!I8s")

class NTP3(NTP4): pass
class NTP2(NTP4): pass
class NTP1(NTP4): pass

def NTP(pktt):
    """Wrapper function to select correct NTP object"""
    unpack = pktt.unpack
    # Check NTP version without consuming any bytes from the Unpack object
    offset = unpack.tell()
    tmp = unpack.unpack(1, "!B")[0]
    unpack.seek(offset)
    version = (tmp>>3)&0x7
    if version == 4:
        return NTP4(pktt)
    elif version == 3:
        return NTP3(pktt)
    elif version == 2:
        return NTP2(pktt)
    elif version == 1:
        return NTP1(pktt)



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Record module

Provides the object for a record and the string representation of the record
in a tcpdump trace file.
"""
import time
import struct
import nfstest_config as c
from baseobj import BaseObj

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "2.1"

FRAME  = 0
INDEX  = 1
TSTAMP = 1

class Record(BaseObj):
    """Record object

       Usage:
           from packet.record import Record

           x = Record(pktt, data)

       Object definition:

       Record(
           frame       = int,   # Frame number
           index       = int,   # Packet number
           seconds     = int,   # Seconds
           usecs       = int,   # Microseconds
           length_inc  = int,   # Number of bytes included in trace
           length_orig = int,   # Number of bytes in packet
           secs        = float, # Absolute seconds including microseconds
           rsecs       = float, # Seconds relative to first packet
       )
    """
    # Class attributes
    _attrlist = ("frame", "index", "seconds", "usecs",
                 "length_inc", "length_orig", "secs", "rsecs")

    def __init__(self, pktt, data):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
           data:
               Raw packet data for this layer.
        """
        # Decode record header
        ulist = struct.unpack(pktt.header_rec, data)
        self.frame       = pktt.frame
        self.index       = pktt.index
        self.seconds     = ulist[0]
        self.usecs       = ulist[1]
        self.length_inc  = ulist[2]
        self.length_orig = ulist[3]
        pktt.pkt.record = self
        # Seconds + microseconds
        self.secs = float(self.seconds) + float(self.usecs)/1000000.0

        if pktt.tstart is None:
            # This is the first packet
            pktt.tstart = self.secs
        # Seconds relative to first packet
        self.rsecs = self.secs - pktt.tstart

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed to display
           either or both the frame or packet numbers and the timestamp:
               '57 2014-03-16 13:42:56.530957 '

           If set to 2 the representation of the object also includes the number
           of bytes on the wire, number of bytes captured and a little bit more
           verbose:
               'frame 57 @ 2014-03-16 13:42:56.530957, 42 bytes on wire, 42 packet bytes'
        """
        idxstr = ""
        tstamp = ""
        rdebug = self.debug_repr()
        if TSTAMP and rdebug in [1,2]:
            tstamp = "%s.%06d" % (time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.seconds)), self.usecs)
        if rdebug == 1:
            if FRAME and INDEX:
                idxstr = "%d,%d " % (self.frame, self.index)
            elif FRAME:
                idxstr = "%d " % self.frame
            elif INDEX:
                idxstr = "%d " % self.index
            out = "%s%s " % (idxstr, tstamp)
        elif rdebug == 2:
            if FRAME and INDEX:
                idxstr = "frame %d,%d @ " % (self.frame, self.index)
            elif FRAME:
                idxstr = "frame %d @ " % self.frame
            elif INDEX:
                idxstr = "index %d @ " % self.index
            out = "%s%s, %d bytes on wire, %d packet bytes" % (idxstr, tstamp, self.length_inc, self.length_orig)
        else:
            out = BaseObj.__str__(self)
        return out



#===============================================================================
# Copyright 2016 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
ARP module

Decode ARP and RARP layers.

RFC 826 An Ethernet Address Resolution Protocol
RFC 903 A Reverse Address Resolution Protocol
"""
import arp_const as const
import nfstest_config as c
from packet.utils import *
from baseobj import BaseObj
from ipv6addr import IPv6Addr
from packet.link.macaddr import MacAddr

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2016 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

class arp_oper(Enum):
    """enum arp_oper"""
    _enumdict = const.arp_oper

class ARP(BaseObj):
    """ARP object

       Usage:
           from packet.internet.arp import ARP

           x = ARP(pktt)

       Object definition:

       ARP(
           htype = int,    # Hardware type
           ptype = int,    # Protocol type
           hlen  = int,    # Byte length for each hardware address
           plen  = int,    # Byte length for each protocol address
           oper  = int,    # Opcode
           sha   = string, # Hardware address of sender of this packet
           spa   = string, # Protocol address of sender of this packet
           tha   = string, # Hardware address of target of this packet
           tpa   = string, # Protocol address of target of this packet
       )
    """
    # Class attributes
    _attrlist = ("htype", "ptype", "hlen", "plen", "oper",
                 "sha", "spa", "tha", "tpa")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        unpack = pktt.unpack
        ulist = unpack.unpack(8, "!HHBBH")
        self.htype = ulist[0]
        self.ptype = ulist[1]
        self.hlen  = ulist[2]
        self.plen  = ulist[3]
        self.oper  = arp_oper(ulist[4])
        self.sha   = self._getha(unpack)
        self.spa   = self._getpa(unpack)
        self.tha   = self._getha(unpack)
        self.tpa   = self._getpa(unpack)

        if self.oper == const.REQUEST:
            self._strfmt1 = "ARP {4} {8}"
            self._strfmt2 = "{4}: Who is {8}? Tell {6}"
        elif self.oper == const.REPLY:
            self._strfmt1 = "ARP {4} {5}"
            self._strfmt2 = "{4}: {6} is {5}"
        elif self.oper == const.RARP_REQUEST:
            self._strfmt1 = "RARP {4} {7}"
            self._strfmt2 = "{4}: Who is {7}? Tell {5}"
        elif self.oper == const.RARP_REPLY:
            self._strfmt1 = "RARP {4} {8}"
            self._strfmt2 = "{4}: {7} is {8}"

        # Set packet layer
        setattr(pktt.pkt, self.__class__.__name__.lower(), self)

    def _getha(self, unpack):
        """Get hardware address"""
        ret = None
        if self.htype == const.HTYPE_ETHERNET:
            ret = MacAddr(unpack.read(6).encode('hex'))
        else:
            ret = unpack.read(self.hlen)
        return ret

    def _getpa(self, unpack):
        """Get protocol address"""
        ret = None
        if self.ptype == const.PTYPE_IPV4:
            ret = "%d.%d.%d.%d" % unpack.unpack(4, "!4B")
        elif self.ptype == const.PTYPE_IPV6:
            ret = IPv6Addr(unpack.read(16).encode('hex'))
        else:
            ret = unpack.read(self.plen)
        return ret

class RARP(ARP): pass



#===============================================================================
# Copyright 2015 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
KRB5 constants module

Provide constant values and mapping dictionaries for the KRB5 layer.

RFC 4120 The Kerberos Network Authentication Service (V5)
RFC 6113 A Generalized Framework for Kerberos Pre-Authentication
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2015 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

# KDCOptions
kdc_options = {
     0 : "reserved",
     1 : "forwardable",
     2 : "forwarded",
     3 : "proxiable",
     4 : "proxy",
     5 : "allow_postdate",
     6 : "postdated",
     8 : "renewable",
    11 : "opt_hardware_auth",
    14 : "constrained_delegation",
    15 : "canonicalize",
    26 : "disable_transited_check",
    27 : "renewable_ok",
    28 : "enc_tkt_in_skey",
    30 : "renew",
    31 : "validate",
}

# APOptions
ap_options = {
    0 : "reserved",
    1 : "use_session_key",
    2 : "mutual_required",
}

# Enum krb5_application
Ticket         = 1   # PDU
Authenticator  = 2   # non-PDU
EncTicketPart  = 3   # non-PDU
AS_REQ         = 10  # PDU
AS_REP         = 11  # PDU
TGS_REQ        = 12  # PDU
TGS_REP        = 13  # PDU
AP_REQ         = 14  # PDU
AP_REP         = 15  # PDU
RESERVED16     = 16  # TGT-REQ (for user-to-user)
RESERVED17     = 17  # TGT-REP (for user-to-user)
KRB_SAFE       = 20  # PDU
KRB_PRIV       = 21  # PDU
KRB_CRED       = 22  # PDU
EncASRepPart   = 25  # non-PDU
EncTGSRepPart  = 26  # non-PDU
EncApRepPart   = 27  # non-PDU
EncKrbPrivPart = 28  # non-PDU
EncKrbCredPart = 29  # non-PDU
KRB_ERROR      = 30  # PDU

krb5_application = {
     1 : "Ticket",
     2 : "Authenticator",
     3 : "EncTicketPart",
    10 : "AS-REQ",
    11 : "AS-REP",
    12 : "TGS-REQ",
    13 : "TGS-REP",
    14 : "AP-REQ",
    15 : "AP-REP",
    16 : "RESERVED16",
    17 : "RESERVED17",
    20 : "KRB-SAFE",
    21 : "KRB-PRIV",
    22 : "KRB-CRED",
    25 : "EncASRepPart",
    26 : "EncTGSRepPart",
    27 : "EncApRepPart",
    28 : "EncKrbPrivPart",
    29 : "EncKrbCredPart",
    30 : "KRB-ERROR",
}

# Enum krb5_principal
UNKNOWN        = 0   # Name type not known
PRINCIPAL      = 1   # Just the name of the principal as in DCE, or for users
SRV_INST       = 2   # Service and other unique instance (krbtgt)
SRV_HST        = 3   # Service with host name as instance (telnet, rcommands)
SRV_XHST       = 4   # Service with host as remaining components
UID            = 5   # Unique ID
X500_PRINCIPAL = 6   # Encoded X.509 Distinguished name [RFC2253]
SMTP_NAME      = 7   # Name in form of SMTP email name (e.g., user@example.com)
ENTERPRISE     = 10  # Enterprise name - may be mapped to principal name

krb5_principal = {
     0 : "UNKNOWN",
     1 : "PRINCIPAL",
     2 : "SRV-INST",
     3 : "SRV-HST",
     4 : "SRV-XHST",
     5 : "UID",
     6 : "X500-PRINCIPAL",
     7 : "SMTP-NAME",
    10 : "ENTERPRISE",
}

# Enum krb5_patype
PA_TGS_REQ                 = 1    # [RFC4120]
PA_ENC_TIMESTAMP           = 2    # [RFC4120]
PA_PW_SALT                 = 3    # [RFC4120]
PA_ENC_UNIX_TIME           = 5    # (deprecated) [RFC4120]
PA_SANDIA_SECUREID         = 6    # [RFC4120]
PA_SESAME                  = 7    # [RFC4120]
PA_OSF_DCE                 = 8    # [RFC4120]
PA_CYBERSAFE_SECUREID      = 9    # [RFC4120]
PA_AFS3_SALT               = 10   # [RFC4120][RFC3961]
PA_ETYPE_INFO              = 11   # [RFC4120]
PA_SAM_CHALLENGE           = 12   # [draft-ietf-cat-kerberos-passwords-04]
PA_SAM_RESPONSE            = 13   # [draft-ietf-cat-kerberos-passwords-04]
PA_PK_AS_REQ_OLD           = 14   # [draft-ietf-cat-kerberos-pk-init-09]
PA_PK_AS_REP_OLD           = 15   # [draft-ietf-cat-kerberos-pk-init-09]
PA_PK_AS_REQ               = 16   # [RFC4556]
PA_PK_AS_REP               = 17   # [RFC4556]
PA_PK_OCSP_RESPONSE        = 18   # [RFC4557]
PA_ETYPE_INFO2             = 19   # [RFC4120]
PA_USE_SPECIFIED_KVNO      = 20   # [RFC4120]
PA_SVR_REFERRAL_INFO       = 20   # [RFC6806]
PA_SAM_REDIRECT            = 21   # [draft-ietf-krb-wg-kerberos-sam-03]
PA_GET_FROM_TYPED_DATA     = 22   # [(embedded in typed data)][RFC4120]
TD_PADATA                  = 22   # [(embeds padata)][RFC4120]
PA_SAM_ETYPE_INFO          = 23   # [(sam/otp)][draft-ietf-krb-wg-kerberos-sam-03]
PA_ALT_PRINC               = 24   # [draft-ietf-krb-wg-hw-auth-04]
PA_SERVER_REFERRAL         = 25   # [draft-ietf-krb-wg-kerberos-referrals-11]
PA_SAM_CHALLENGE2          = 30   # [draft-ietf-krb-wg-kerberos-sam-03]
PA_SAM_RESPONSE2           = 31   # [draft-ietf-krb-wg-kerberos-sam-03]
PA_EXTRA_TGT               = 41   # [Reserved extra TGT][RFC6113]
TD_PKINIT_CMS_CERTIFICATES = 101  # [RFC4556]
TD_KRB_PRINCIPAL           = 102  # [PrincipalName][RFC6113]
TD_KRB_REALM               = 103  # [Realm][RFC6113]
TD_TRUSTED_CERTIFIERS      = 104  # [RFC4556]
TD_CERTIFICATE_INDEX       = 105  # [RFC4556]
TD_APP_DEFINED_ERROR       = 106  # [Application specific][RFC6113]
TD_REQ_NONCE               = 107  # [INTEGER][RFC6113]
TD_REQ_SEQ                 = 108  # [INTEGER][RFC6113]
TD_DH_PARAMETERS           = 109  # [RFC4556]
TD_CMS_DIGEST_ALGORITHMS   = 111  # [draft-ietf-krb-wg-pkinit-alg-agility]
TD_CERT_DIGEST_ALGORITHMS  = 112  # [draft-ietf-krb-wg-pkinit-alg-agility]
PA_PAC_REQUEST             = 128  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_FOR_USER                = 129  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_FOR_X509_USER           = 130  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_FOR_CHECK_DUPS          = 131  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_AS_CHECKSUM             = 132  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_FX_COOKIE               = 133  # [RFC6113]
PA_AUTHENTICATION_SET      = 134  # [RFC6113]
PA_AUTH_SET_SELECTED       = 135  # [RFC6113]
PA_FX_FAST                 = 136  # [RFC6113]
PA_FX_ERROR                = 137  # [RFC6113]
PA_ENCRYPTED_CHALLENGE     = 138  # [RFC6113]
PA_OTP_CHALLENGE           = 141  # [RFC6560]
PA_OTP_REQUEST             = 142  # [RFC6560]
PA_OTP_CONFIRM             = 143  # (OBSOLETED) [RFC6560]
PA_OTP_PIN_CHANGE          = 144  # [RFC6560]
PA_EPAK_AS_REQ             = 145  # [RFC6113]
PA_EPAK_AS_REP             = 146  # [RFC6113]
PA_PKINIT_KX               = 147  # [RFC6112]
PA_PKU2U_NAME              = 148  # [draft-zhu-pku2u]
PA_REQ_ENC_PA_REP          = 149  # [RFC6806]
PA_AS_FRESHNESS            = 150  # [draft-ietf-kitten-pkinit-freshness]
PA_SUPPORTED_ETYPES        = 165  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]
PA_EXTENDED_ERROR          = 166  # [MSKILE][http://msdn2.microsoft.com/en-us/library/cc206927.aspx]

krb5_patype = {
      1 : "PA-TGS-REQ",
      2 : "PA-ENC-TIMESTAMP",
      3 : "PA-PW-SALT",
      5 : "PA-ENC-UNIX-TIME",
      6 : "PA-SANDIA-SECUREID",
      7 : "PA-SESAME",
      8 : "PA-OSF-DCE",
      9 : "PA-CYBERSAFE-SECUREID",
     10 : "PA-AFS3-SALT",
     11 : "PA-ETYPE-INFO",
     12 : "PA-SAM-CHALLENGE",
     13 : "PA-SAM-RESPONSE",
     14 : "PA-PK-AS-REQ_OLD",
     15 : "PA-PK-AS-REP_OLD",
     16 : "PA-PK-AS-REQ",
     17 : "PA-PK-AS-REP",
     18 : "PA-PK-OCSP-RESPONSE",
     19 : "PA-ETYPE-INFO2",
     20 : "PA-USE-SPECIFIED-KVNO",
     20 : "PA-SVR-REFERRAL-INFO",
     21 : "PA-SAM-REDIRECT",
     22 : "PA-GET-FROM-TYPED-DATA",
     22 : "TD-PADATA",
     23 : "PA-SAM-ETYPE-INFO",
     24 : "PA-ALT-PRINC",
     25 : "PA-SERVER-REFERRAL",
     30 : "PA-SAM-CHALLENGE2",
     31 : "PA-SAM-RESPONSE2",
     41 : "PA-EXTRA-TGT",
    101 : "TD-PKINIT-CMS-CERTIFICATES",
    102 : "TD-KRB-PRINCIPAL",
    103 : "TD-KRB-REALM",
    104 : "TD-TRUSTED-CERTIFIERS",
    105 : "TD-CERTIFICATE-INDEX",
    106 : "TD-APP-DEFINED-ERROR",
    107 : "TD-REQ-NONCE",
    108 : "TD-REQ-SEQ",
    109 : "TD_DH_PARAMETERS",
    111 : "TD-CMS-DIGEST-ALGORITHMS",
    112 : "TD-CERT-DIGEST-ALGORITHMS",
    128 : "PA-PAC-REQUEST",
    129 : "PA-FOR_USER",
    130 : "PA-FOR-X509-USER",
    131 : "PA-FOR-CHECK_DUPS",
    132 : "PA-AS-CHECKSUM",
    133 : "PA-FX-COOKIE",
    134 : "PA-AUTHENTICATION-SET",
    135 : "PA-AUTH-SET-SELECTED",
    136 : "PA-FX-FAST",
    137 : "PA-FX-ERROR",
    138 : "PA-ENCRYPTED-CHALLENGE",
    141 : "PA-OTP-CHALLENGE",
    142 : "PA-OTP-REQUEST",
    143 : "PA-OTP-CONFIRM",
    144 : "PA-OTP-PIN-CHANGE",
    145 : "PA-EPAK-AS-REQ",
    146 : "PA-EPAK-AS-REP",
    147 : "PA_PKINIT_KX",
    148 : "PA_PKU2U_NAME",
    149 : "PA-REQ-ENC-PA-REP",
    150 : "PA_AS_FRESHNESS",
    165 : "PA-SUPPORTED-ETYPES",
    166 : "PA-EXTENDED_ERROR",
}

# Enum krb5_addrtype
IPv4            = 2
Directional     = 3
ChaosNet        = 5
XNS             = 6
ISO             = 7
DECNET_Phase_IV = 12
AppleTalk_DDP   = 16
NetBios         = 20
IPv6            = 24

krb5_addrtype = {
     2 : "IPv4",
     3 : "Directional",
     5 : "ChaosNet",
     6 : "XNS",
     7 : "ISO",
    12 : "DECNET-Phase-IV",
    16 : "AppleTalk-DDP",
    20 : "NetBios",
    24 : "IPv6",
}

# Enum krb5_adtype
AD_IF_RELEVANT                    = 1
AD_INTENDED_FOR_SERVER            = 2
AD_INTENDED_FOR_APPLICATION_CLASS = 3
AD_KDC_ISSUED                     = 4
AD_AND_OR                         = 5
AD_MANDATORY_TICKET_EXTENSIONS    = 6
AD_IN_TICKET_EXTENSIONS           = 7
AD_MANDATORY_FOR_KDC              = 8
OSF_DCE                           = 64
SESAME                            = 65
AD_OSF_DCE_PKI_CERTID             = 66
AD_WIN2K_PAC                      = 128
AD_ETYPE_NEGOTIATION              = 129

krb5_adtype = {
      1 : "AD-IF-RELEVANT",
      2 : "AD-INTENDED-FOR-SERVER",
      3 : "AD-INTENDED-FOR-APPLICATION-CLASS",
      4 : "AD-KDC-ISSUED",
      5 : "AD-AND-OR",
      6 : "AD-MANDATORY-TICKET-EXTENSIONS",
      7 : "AD-IN-TICKET-EXTENSIONS",
      8 : "AD-MANDATORY-FOR-KDC",
     64 : "OSF-DCE",
     65 : "SESAME",
     66 : "AD-OSF-DCE-PKI-CERTID",
    128 : "AD-WIN2K-PAC",
    129 : "AD-ETYPE-NEGOTIATION",
}

# Enum krb5_etype
des_cbc_crc                  = 1   # [RFC3961]
des_cbc_md4                  = 2   # [RFC3961]
des_cbc_md5                  = 3   # [RFC3961]
des3_cbc_md5                 = 5
des3_cbc_sha1                = 7
dsaWithSHA1_CmsOID           = 9   # [RFC4556]
md5WithRSAEncryption_CmsOID  = 10  # [RFC4556]
sha1WithRSAEncryption_CmsOID = 11  # [RFC4556]
rc2CBC_EnvOID                = 12  # [RFC4556]
rsaEncryption_EnvOID         = 13  # [RFC4556][from PKCS#1 v1.5]]
rsaES_OAEP_ENV_OID           = 14  # [RFC4556][from PKCS#1 v2.0]]
des_ede3_cbc_Env_OID         = 15  # [RFC4556]
des3_cbc_sha1_kd             = 16  # [RFC3961]
aes128_cts_hmac_sha1_96      = 17  # [RFC3962]
aes256_cts_hmac_sha1_96      = 18  # [RFC3962]
rc4_hmac                     = 23  # [RFC4757]
rc4_hmac_exp                 = 24  # [RFC4757]
camellia128_cts_cmac         = 25  # [RFC6803]
camellia256_cts_cmac         = 26  # [RFC6803]
subkey_keymaterial           = 65  # [(opaque; PacketCable)]

krb5_etype = {
     1 : "des-cbc-crc",
     2 : "des-cbc-md4",
     3 : "des-cbc-md5",
     5 : "des3-cbc-md5",
     7 : "des3-cbc-sha1",
     9 : "dsaWithSHA1-CmsOID",
    10 : "md5WithRSAEncryption-CmsOID",
    11 : "sha1WithRSAEncryption-CmsOID",
    12 : "rc2CBC-EnvOID",
    13 : "rsaEncryption-EnvOID",
    14 : "rsaES-OAEP-ENV-OID",
    15 : "des-ede3-cbc-Env-OID",
    16 : "des3-cbc-sha1-kd",
    17 : "aes128-cts-hmac-sha1-96",
    18 : "aes256-cts-hmac-sha1-96",
    23 : "rc4-hmac",
    24 : "rc4-hmac-exp",
    25 : "camellia128-cts-cmac",
    26 : "camellia256-cts-cmac",
    65 : "subkey-keymaterial",
}

# Enum krb5_ctype
CRC32               = 1   # Checksum size:4  [RFC3961]
rsa_md4             = 2   # Checksum size:16 [RFC3961]
rsa_md4_des         = 3   # Checksum size:24 [RFC3961]
des_mac             = 4   # Checksum size:16 [RFC3961]
des_mac_k           = 5   # Checksum size:8  [RFC3961]
rsa_md4_des_k       = 6   # Checksum size:16 [RFC3961]
rsa_md5             = 7   # Checksum size:16 [RFC3961]
rsa_md5_des         = 8   # Checksum size:24 [RFC3961]
rsa_md5_des3        = 9   # Checksum size:24
sha1                = 10  # Checksum size:20 (unkeyed)
hmac_sha1_des3_kd   = 12  # Checksum size:20 [RFC3961]
hmac_sha1_des3      = 13  # Checksum size:20
sha1                = 14  # Checksum size:20 (unkeyed)
hmac_sha1_96_aes128 = 15  # Checksum size:20 [RFC3962]
hmac_sha1_96_aes256 = 16  # Checksum size:20 [RFC3962]
cmac_camellia128    = 17  # Checksum size:16 [RFC6803]
cmac_camellia256    = 18  # Checksum size:16 [RFC6803]

krb5_ctype = {
     1 : "CRC32",
     2 : "rsa-md4",
     3 : "rsa-md4-des",
     4 : "des-mac",
     5 : "des-mac-k",
     6 : "rsa-md4-des-k",
     7 : "rsa-md5",
     8 : "rsa-md5-des",
     9 : "rsa-md5-des3",
    10 : "sha1",
    12 : "hmac-sha1-des3-kd",
    13 : "hmac-sha1-des3",
    14 : "sha1",
    15 : "hmac-sha1-96-aes128",
    16 : "hmac-sha1-96-aes256",
    17 : "cmac-camellia128",
    18 : "cmac-camellia256",
}

# Enum krb5_fatype
RESERVED                 = 0
FX_FAST_ARMOR_AP_REQUEST = 1  # Ticket armor using an ap-req

krb5_fatype = {
    0 : "RESERVED",
    1 : "FX_FAST_ARMOR_AP_REQUEST",
}

# Enum krb5_status
KDC_OK                                = 0   # No error
KDC_ERR_NAME_EXP                      = 1   # Client's entry in database has expired
KDC_ERR_SERVICE_EXP                   = 2   # Server's entry in database has expired
KDC_ERR_BAD_PVNO                      = 3   # Requested protocol version number not supported
KDC_ERR_C_OLD_MAST_KVNO               = 4   # Client's key encrypted in old master key
KDC_ERR_S_OLD_MAST_KVNO               = 5   # Server's key encrypted in old master key
KDC_ERR_C_PRINCIPAL_UNKNOWN           = 6   # Client not found in Kerberos database
KDC_ERR_S_PRINCIPAL_UNKNOWN           = 7   # Server not found in Kerberos database
KDC_ERR_PRINCIPAL_NOT_UNIQUE          = 8   # Multiple principal entries in database
KDC_ERR_NULL_KEY                      = 9   # The client or server has a null key
KDC_ERR_CANNOT_POSTDATE               = 10  # Ticket not eligible for postdating
KDC_ERR_NEVER_VALID                   = 11  # Requested starttime is later than end time
KDC_ERR_POLICY                        = 12  # KDC policy rejects request
KDC_ERR_BADOPTION                     = 13  # KDC cannot accommodate requested option
KDC_ERR_ETYPE_NOSUPP                  = 14  # KDC has no support for encryption type
KDC_ERR_SUMTYPE_NOSUPP                = 15  # KDC has no support for checksum type
KDC_ERR_PADATA_TYPE_NOSUPP            = 16  # KDC has no support for padata type
KDC_ERR_TRTYPE_NOSUPP                 = 17  # KDC has no support for transited type
KDC_ERR_CLIENT_REVOKED                = 18  # Clients credentials have been revoked
KDC_ERR_SERVICE_REVOKED               = 19  # Credentials for server have been revoked
KDC_ERR_TGT_REVOKED                   = 20  # TGT has been revoked
KDC_ERR_CLIENT_NOTYET                 = 21  # Client not yet valid; try again later
KDC_ERR_SERVICE_NOTYET                = 22  # Server not yet valid; try again later
KDC_ERR_KEY_EXPIRED                   = 23  # Password has expired; change password to reset
KDC_ERR_PREAUTH_FAILED                = 24  # Pre-authentication information was invalid
KDC_ERR_PREAUTH_REQUIRED              = 25  # Additional pre-authentication required
KDC_ERR_SERVER_NOMATCH                = 26  # Requested server and ticket don't match
KDC_ERR_MUST_USE_USER2USER            = 27  # Server principal valid for user2user only
KDC_ERR_PATH_NOT_ACCEPTED             = 28  # KDC Policy rejects transited path
KDC_ERR_SVC_UNAVAILABLE               = 29  # A service is not available
KRB_AP_ERR_BAD_INTEGRITY              = 31  # Integrity check on decrypted field failed
KRB_AP_ERR_TKT_EXPIRED                = 32  # Ticket expired
KRB_AP_ERR_TKT_NYV                    = 33  # Ticket not yet valid
KRB_AP_ERR_REPEAT                     = 34  # Request is a replay
KRB_AP_ERR_NOT_US                     = 35  # The ticket isn't for us
KRB_AP_ERR_BADMATCH                   = 36  # Ticket and authenticator don't match
KRB_AP_ERR_SKEW                       = 37  # Clock skew too great
KRB_AP_ERR_BADADDR                    = 38  # Incorrect net address
KRB_AP_ERR_BADVERSION                 = 39  # Protocol version mismatch
KRB_AP_ERR_MSG_TYPE                   = 40  # Invalid msg type
KRB_AP_ERR_MODIFIED                   = 41  # Message stream modified
KRB_AP_ERR_BADORDER                   = 42  # Message out of order
KRB_AP_ERR_BADKEYVER                  = 44  # Specified version of key is not available
KRB_AP_ERR_NOKEY                      = 45  # Service key not available
KRB_AP_ERR_MUT_FAIL                   = 46  # Mutual authentication failed
KRB_AP_ERR_BADDIRECTION               = 47  # Incorrect message direction
KRB_AP_ERR_METHOD                     = 48  # Alternative authentication method required
KRB_AP_ERR_BADSEQ                     = 49  # Incorrect sequence number in message
KRB_AP_ERR_INAPP_CKSUM                = 50  # Inappropriate type of checksum in message
KRB_AP_ERR_PATH_NOT_ACCEPTED          = 51  # Policy rejects transited path
KRB_ERR_RESPONSE_TOO_BIG              = 52  # Response too big for UDP; retry with TCP
KRB_ERR_GENERIC                       = 60  # Generic error (description in e-text)
KRB_ERR_FIELD_TOOLONG                 = 61  # Field is too long for this implementation
KDC_ERR_CLIENT_NOT_TRUSTED            = 62  # Reserved for PKINIT
KDC_ERR_KDC_NOT_TRUSTED               = 63  # Reserved for PKINIT
KDC_ERR_INVALID_SIG                   = 64  # Reserved for PKINIT
KDC_ERR_KEY_TOO_WEAK                  = 65  # Reserved for PKINIT
KDC_ERR_CERTIFICATE_MISMATCH          = 66  # Reserved for PKINIT
KRB_AP_ERR_NO_TGT                     = 67  # No TGT available to validate USER-TO-USER
KDC_ERR_WRONG_REALM                   = 68  # Reserved for future use
KRB_AP_ERR_USER_TO_USER_REQUIRED      = 69  # Ticket must be for USER-TO-USER
KDC_ERR_CANT_VERIFY_CERTIFICATE       = 70  # Reserved for PKINIT
KDC_ERR_INVALID_CERTIFICATE           = 71  # Reserved for PKINIT
KDC_ERR_REVOKED_CERTIFICATE           = 72  # Reserved for PKINIT
KDC_ERR_REVOCATION_STATUS_UNKNOWN     = 73  # Reserved for PKINIT
KDC_ERR_REVOCATION_STATUS_UNAVAILABLE = 74  # Reserved for PKINIT
KDC_ERR_CLIENT_NAME_MISMATCH          = 75  # Reserved for PKINIT
KDC_ERR_KDC_NAME_MISMATCH             = 76  # Reserved for PKINIT

krb5_status = {
     0 : "KDC_OK",
     1 : "KDC_ERR_NAME_EXP",
     2 : "KDC_ERR_SERVICE_EXP",
     3 : "KDC_ERR_BAD_PVNO",
     4 : "KDC_ERR_C_OLD_MAST_KVNO",
     5 : "KDC_ERR_S_OLD_MAST_KVNO",
     6 : "KDC_ERR_C_PRINCIPAL_UNKNOWN",
     7 : "KDC_ERR_S_PRINCIPAL_UNKNOWN",
     8 : "KDC_ERR_PRINCIPAL_NOT_UNIQUE",
     9 : "KDC_ERR_NULL_KEY",
    10 : "KDC_ERR_CANNOT_POSTDATE",
    11 : "KDC_ERR_NEVER_VALID",
    12 : "KDC_ERR_POLICY",
    13 : "KDC_ERR_BADOPTION",
    14 : "KDC_ERR_ETYPE_NOSUPP",
    15 : "KDC_ERR_SUMTYPE_NOSUPP",
    16 : "KDC_ERR_PADATA_TYPE_NOSUPP",
    17 : "KDC_ERR_TRTYPE_NOSUPP",
    18 : "KDC_ERR_CLIENT_REVOKED",
    19 : "KDC_ERR_SERVICE_REVOKED",
    20 : "KDC_ERR_TGT_REVOKED",
    21 : "KDC_ERR_CLIENT_NOTYET",
    22 : "KDC_ERR_SERVICE_NOTYET",
    23 : "KDC_ERR_KEY_EXPIRED",
    24 : "KDC_ERR_PREAUTH_FAILED",
    25 : "KDC_ERR_PREAUTH_REQUIRED",
    26 : "KDC_ERR_SERVER_NOMATCH",
    27 : "KDC_ERR_MUST_USE_USER2USER",
    28 : "KDC_ERR_PATH_NOT_ACCEPTED",
    29 : "KDC_ERR_SVC_UNAVAILABLE",
    31 : "KRB_AP_ERR_BAD_INTEGRITY",
    32 : "KRB_AP_ERR_TKT_EXPIRED",
    33 : "KRB_AP_ERR_TKT_NYV",
    34 : "KRB_AP_ERR_REPEAT",
    35 : "KRB_AP_ERR_NOT_US",
    36 : "KRB_AP_ERR_BADMATCH",
    37 : "KRB_AP_ERR_SKEW",
    38 : "KRB_AP_ERR_BADADDR",
    39 : "KRB_AP_ERR_BADVERSION",
    40 : "KRB_AP_ERR_MSG_TYPE",
    41 : "KRB_AP_ERR_MODIFIED",
    42 : "KRB_AP_ERR_BADORDER",
    44 : "KRB_AP_ERR_BADKEYVER",
    45 : "KRB_AP_ERR_NOKEY",
    46 : "KRB_AP_ERR_MUT_FAIL",
    47 : "KRB_AP_ERR_BADDIRECTION",
    48 : "KRB_AP_ERR_METHOD",
    49 : "KRB_AP_ERR_BADSEQ",
    50 : "KRB_AP_ERR_INAPP_CKSUM",
    51 : "KRB_AP_ERR_PATH_NOT_ACCEPTED",
    52 : "KRB_ERR_RESPONSE_TOO_BIG",
    60 : "KRB_ERR_GENERIC",
    61 : "KRB_ERR_FIELD_TOOLONG",
    62 : "KDC_ERR_CLIENT_NOT_TRUSTED",
    63 : "KDC_ERR_KDC_NOT_TRUSTED",
    64 : "KDC_ERR_INVALID_SIG",
    65 : "KDC_ERR_KEY_TOO_WEAK",
    66 : "KDC_ERR_CERTIFICATE_MISMATCH",
    67 : "KRB_AP_ERR_NO_TGT",
    68 : "KDC_ERR_WRONG_REALM",
    69 : "KRB_AP_ERR_USER_TO_USER_REQUIRED",
    70 : "KDC_ERR_CANT_VERIFY_CERTIFICATE",
    71 : "KDC_ERR_INVALID_CERTIFICATE",
    72 : "KDC_ERR_REVOKED_CERTIFICATE",
    73 : "KDC_ERR_REVOCATION_STATUS_UNKNOWN",
    74 : "KDC_ERR_REVOCATION_STATUS_UNAVAILABLE",
    75 : "KDC_ERR_CLIENT_NAME_MISMATCH",
    76 : "KDC_ERR_KDC_NAME_MISMATCH",
}



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
DNS constants module

Provide constant values and mapping dictionaries for the DNS layer.

RFC 1035 Domain Names - Implementation and Specification
RFC 2671 Extension Mechanisms for DNS (EDNS0)
RFC 4034 Resource Records for the DNS Security Extensions
RFC 4035 Protocol Modifications for the DNS Security Extensions
RFC 4255 Using DNS to Securely Publish Secure Shell (SSH) Key Fingerprints
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.0"

# Enum dns_query
QUERY = 0
REPLY = 1

dns_query = {
    0 : "QUERY",
    1 : "REPLY",
}

# Enum dns_opcode
QUERY  = 0
IQUERY = 1
STATUS = 2
NOTIFY = 4
UPDATE = 5

dns_opcode = {
    0 : "QUERY",
    1 : "IQUERY",
    2 : "STATUS",
    4 : "NOTIFY",
    5 : "UPDATE",
}

# Enum dns_rcode
NOERROR          = 0   # No Error [RFC1035]
DNSERR_FORMERR   = 1   # Format Error [RFC1035]
DNSERR_SERVFAIL  = 2   # Server Failure [RFC1035]
DNSERR_NXDOMAIN  = 3   # Non-Existent Domain [RFC1035]
DNSERR_NOTIMP    = 4   # Not Implemented [RFC1035]
DNSERR_REFUSED   = 5   # Query Refused [RFC1035]
DNSERR_YXDOMAIN  = 6   # Name Exists when it should not [RFC2136][RFC6672]
DNSERR_YXRRSET   = 7   # RR Set Exists when it should not [RFC2136]
DNSERR_NXRRSET   = 8   # RR Set that should exist does not [RFC2136]
DNSERR_NOTAUTH   = 9   # Server Not Authoritative for zone [RFC2136]
                       # Not Authorized [RFC2845]
DNSERR_NOTZONE   = 10  # Name not contained in zone [RFC2136]
DNSERR_BADVERS   = 16  # Bad OPT Version [RFC6891]
                       # TSIG Signature Failure [RFC2845]
DNSERR_BADKEY    = 17  # Key not recognized [RFC2845]
DNSERR_BADTIME   = 18  # Signature out of time window [RFC2845]
DNSERR_BADMODE   = 19  # Bad TKEY Mode [RFC2930]
DNSERR_BADNAME   = 20  # Duplicate key name [RFC2930]
DNSERR_BADALG    = 21  # Algorithm not supported [RFC2930]
DNSERR_BADTRUNC  = 22  # Bad Truncation [RFC4635]
DNSERR_BADCOOKIE = 23  # Bad/missing Server Cookie [RFC7873]

dns_rcode = {
     0 : "NOERROR",
     1 : "DNSERR_FORMERR",
     2 : "DNSERR_SERVFAIL",
     3 : "DNSERR_NXDOMAIN",
     4 : "DNSERR_NOTIMP",
     5 : "DNSERR_REFUSED",
     6 : "DNSERR_YXDOMAIN",
     7 : "DNSERR_YXRRSET",
     8 : "DNSERR_NXRRSET",
     9 : "DNSERR_NOTAUTH",
    10 : "DNSERR_NOTZONE",
    16 : "DNSERR_BADVERS",
    17 : "DNSERR_BADKEY",
    18 : "DNSERR_BADTIME",
    19 : "DNSERR_BADMODE",
    20 : "DNSERR_BADNAME",
    21 : "DNSERR_BADALG",
    22 : "DNSERR_BADTRUNC",
    23 : "DNSERR_BADCOOKIE",
}

# Enum dns_type
A          = 1      # Host address
NS         = 2      # Authoritative name server
MD         = 3      # Mail destination (Obsolete - use MX)
MF         = 4      # Mail forwarder (Obsolete - use MX)
CNAME      = 5      # Canonical name for an alias
SOA        = 6      # Marks the start of a zone of authority
MB         = 7      # Mailbox domain name (EXPERIMENTAL)
MG         = 8      # Mail group member (EXPERIMENTAL)
MR         = 9      # Mail rename domain name (EXPERIMENTAL)
NULL       = 10     # Null RR (EXPERIMENTAL)
WKS        = 11     # Well known service description
PTR        = 12     # Domain name pointer
HINFO      = 13     # Host information
MINFO      = 14     # Mailbox or mail list information
MX         = 15     # Mail exchange
TXT        = 16     # Text strings
RP         = 17     # Responsible Person [RFC1183]
AFSDB      = 18     # AFS Data Base location [RFC1183][RFC5864]
X25        = 19     # X.25 PSDN address [RFC1183]
ISDN       = 20     # ISDN address [RFC1183]
RT         = 21     # Route Through [RFC1183]
NSAP       = 22     # NSAP address, NSAP style A record [RFC1706]
NSAPPTR    = 23     # Domain name pointer, NSAP style [RFC1348][RFC1637][RFC1706]
SIG        = 24     # Security signature [RFC4034][RFC3755][RFC2535][RFC2536][RFC2537][RFC2931][RFC3110][RFC3008]
KEY        = 25     # Security key [RFC4034][RFC3755][RFC2535][RFC2536][RFC2537][RFC2539][RFC3008][RFC3110]
PX         = 26     # X.400 mail mapping information [RFC2163]
GPOS       = 27     # Geographical Position [RFC1712]
AAAA       = 28     # IPv6 address
LOC        = 29     # Location record
NXT        = 30     # Next Domain (OBSOLETE) [RFC3755][RFC2535]
EID        = 31     # Endpoint Identifier
NIMLOC     = 32     # Nimrod Locator
SRV        = 33     # Service locator
ATMA       = 34     # ATM Address
NAPTR      = 35     # Naming Authority Pointer [RFC2915][RFC2168][RFC3403]
KX         = 36     # Key Exchanger [RFC2230]
CERT       = 37     # CERT [RFC4398]
A6         = 38     # A6 (OBSOLETE - use AAAA) [RFC3226][RFC2874][RFC6563]
DNAME      = 39     # DNAME [RFC6672]
SINK       = 40     # SINK
OPT        = 41     # OPT pseudo-RR [RFC6891][RFC3225][RFC2671]
APL        = 42     # APL [RFC3123]
DS         = 43     # Delegation Signer [RFC4034][RFC3658]
SSHFP      = 44     # Secure shell fingerprint
IPSECKEY   = 45     # IPSECKEY [RFC4025]
RRSIG      = 46     # Resource record digital signature
NSEC       = 47     # NSEC [RFC4034][RFC3755]
DNSKEY     = 48     # DNSKEY [RFC4034][RFC3755]
DHCID      = 49     # DHCID [RFC4701]
NSEC3      = 50     # NSEC3 [RFC5155]
NSEC3PARAM = 51     # NSEC3PARAM [RFC5155]
TLSA       = 52     # TLSA [RFC6698]
SMIMEA     = 53     # S/MIME cert association [draft-ietf-dane-smime]
HIP        = 55     # Host Identity Protocol [RFC5205]
NINFO      = 56     # NINFO [Jim_Reid] NINFO/ninfo-completed-template 2008-01-21
RKEY       = 57     # RKEY [Jim_Reid] RKEY/rkey-completed-template 2008-01-21
TALINK     = 58     # Trust Anchor LINK [Wouter_Wijngaards] TALINK/talink-completed-template 2010-02-17
CDS        = 59     # Child DS [RFC7344] CDS/cds-completed-template 2011-06-06
CDNSKEY    = 60     # DNSKEY(s) the Child wants reflected in DS [RFC7344] 2014-06-16
OPENPGPKEY = 61     # OpenPGP Key [RFC-ietf-dane-openpgpkey-12] OPENPGPKEY/openpgpkey-completed-template 2014-08-12
CSYNC      = 62     # Child-To-Parent Synchronization [RFC7477] 2015-01-27
SPF        = 99     # [RFC7208]
UINFO      = 100    # [IANA-Reserved]
UID        = 101    # [IANA-Reserved]
GID        = 102    # [IANA-Reserved]
UNSPEC     = 103    # [IANA-Reserved]
NID        = 104    # [RFC6742] ILNP/nid-completed-template
L32        = 105    # [RFC6742] ILNP/l32-completed-template
L64        = 106    # [RFC6742] ILNP/l64-completed-template
LP         = 107    # [RFC6742] ILNP/lp-completed-template
EUI48      = 108    # EUI-48 address [RFC7043] EUI48/eui48-completed-template 2013-03-27
EUI64      = 109    # EUI-64 address [RFC7043] EUI64/eui64-completed-template 2013-03-27
TKEY       = 249    # Transaction Key [RFC2930]
TSIG       = 250    # Transaction Signature [RFC2845]
IXFR       = 251    # Incremental transfer [RFC1995]
AXFR       = 252    # Transfer of an entire zone [RFC1035][RFC5936]
MAILB      = 253    # Mailbox-related RRs (MB, MG or MR) [RFC1035]
MAILA      = 254    # Mail agent RRs (OBSOLETE - see MX) [RFC1035]
ANY        = 255    # Request all records
URI        = 256    # URI [RFC7553]
CAA        = 257    # Certification Authority Restriction [RFC6844]
AVC        = 258    # Application Visibility and Control
TA         = 32768  # DNSSEC Trust Authorities
DLV        = 32769  # DNSSEC Lookaside Validation [RFC4431]

dns_type = {
        1 : "A",
        2 : "NS",
        3 : "MD",
        4 : "MF",
        5 : "CNAME",
        6 : "SOA",
        7 : "MB",
        8 : "MG",
        9 : "MR",
       10 : "NULL",
       11 : "WKS",
       12 : "PTR",
       13 : "HINFO",
       14 : "MINFO",
       15 : "MX",
       16 : "TXT",
       17 : "RP",
       18 : "AFSDB",
       19 : "X25",
       20 : "ISDN",
       21 : "RT",
       22 : "NSAP",
       23 : "NSAPPTR",
       24 : "SIG",
       25 : "KEY",
       26 : "PX",
       27 : "GPOS",
       28 : "AAAA",
       29 : "LOC",
       30 : "NXT",
       31 : "EID",
       32 : "NIMLOC",
       33 : "SRV",
       34 : "ATMA",
       35 : "NAPTR",
       36 : "KX",
       37 : "CERT",
       38 : "A6",
       39 : "DNAME",
       40 : "SINK",
       41 : "OPT",
       42 : "APL",
       43 : "DS",
       44 : "SSHFP",
       45 : "IPSECKEY",
       46 : "RRSIG",
       47 : "NSEC",
       48 : "DNSKEY",
       49 : "DHCID",
       50 : "NSEC3",
       51 : "NSEC3PARAM",
       52 : "TLSA",
       53 : "SMIMEA",
       55 : "HIP",
       56 : "NINFO",
       57 : "RKEY",
       58 : "TALINK",
       59 : "CDS",
       60 : "CDNSKEY",
       61 : "OPENPGPKEY",
       62 : "CSYNC",
       99 : "SPF",
      100 : "UINFO",
      101 : "UID",
      102 : "GID",
      103 : "UNSPEC",
      104 : "NID",
      105 : "L32",
      106 : "L64",
      107 : "LP",
      108 : "EUI48",
      109 : "EUI64",
      249 : "TKEY",
      250 : "TSIG",
      251 : "IXFR",
      252 : "AXFR",
      253 : "MAILB",
      254 : "MAILA",
      255 : "ANY",
      256 : "URI",
      257 : "CAA",
      258 : "AVC",
    32768 : "TA",
    32769 : "DLV",
}

# Enum dns_class
IN   = 1    # Internet
CS   = 2    # Chaos
CH   = 3    # Hesiod
HS   = 4    # Internet
NONE = 254  # QCLASS None
ANY  = 255  # QCLASS Any

dns_class = {
      1 : "IN",
      2 : "CS",
      3 : "CH",
      4 : "HS",
    254 : "NONE",
    255 : "ANY",
}

# Enum dns_algorithm
RSA     = 1  # RSA Algorithm [RFC4255]
DSS     = 2  # DSS Algorithm [RFC4255]
ECDSA   = 3  # Elliptic Curve Digital Signature Algorithm [RFC6594]
Ed25519 = 4  # Ed25519 Signature Algorithm [RFC7479]

dns_algorithm = {
    1 : "RSA",
    2 : "DSS",
    3 : "ECDSA",
    4 : "Ed25519",
}

# Enum dns_fptype
SHA1   = 1  # Secure Hash Algorithm 1
SHA256 = 2  # Secure Hash Algorithm 256

dns_fptype = {
    1 : "SHA-1",
    2 : "SHA-256",
}



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
IPv6 module

Decode IP version 6 layer.
Extension headers are not supported.
"""
from ipv4 import IPv4
import nfstest_config as c
from ipv6addr import IPv6Addr
from packet.transport.tcp import TCP
from packet.transport.udp import UDP

# Module constants
__author__    = 'Jorge Mora (%s)' % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = '1.0.4'

class IPv6(IPv4):
    """IPv6 object

       Usage:
           from packet.internet.ipv6 import IPv6

           x = IPv6(pktt)

       Object definition:

       IPv6(
           version       = int,
           traffic_class = int,
           flow_label    = int,
           total_size    = int,
           protocol      = int,
           hop_limit     = int,
           src           = IPv6Addr(),
           dst           = IPv6Addr(),
           data          = string,  # raw data of payload if protocol
                                    # is not supported
       )
    """
    # Class attributes
    _attrlist = ("version", "traffic_class", "flow_label", "total_size",
                 "protocol", "hop_limit", "src", "dst", "data")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        unpack = pktt.unpack
        ulist = unpack.unpack(40, "!IHBB16s16s")
        self.version       = (ulist[0] >> 28)
        self.traffic_class = (ulist[0] >> 20)&0xFF
        self.flow_label    = ulist[0]&0xFFF
        self.total_size    = ulist[1]
        self.protocol      = ulist[2]
        self.hop_limit     = ulist[3]
        self.src           = IPv6Addr(ulist[4].encode('hex'))
        self.dst           = IPv6Addr(ulist[5].encode('hex'))

        pktt.pkt.ip = self

        if self.protocol == 6:
            # Decode TCP
            TCP(pktt)
        elif self.protocol == 17:
            # Decode UDP
            UDP(pktt)
        else:
            self.data = unpack.getbytes()
